#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Clients deployment configuration I/O."""

import pathlib
import typing

import pydantic

from ._exceptions import ClientsDefinitionError
from ._loader import load_configuration

ENTITY_NAME = "clients"

ReleaseStatesValueType = typing.List[str]


class SingularClientDefinition(pydantic.BaseModel):
    """A single client definition."""

    depends_on: typing.Optional[typing.Dict[str, typing.Optional[str]]]
    pseudonym: typing.Optional[str]
    release_states: ReleaseStatesValueType
    system: str


T = typing.TypeVar("T", bound="ClientsDefinition")


ValueType = typing.Dict[str, SingularClientDefinition]


class ClientsDefinition(pydantic.BaseModel):
    """Definition of deployment clients."""

    clients: ValueType

    @pydantic.validator(ENTITY_NAME)
    def check_clients(
        cls: pydantic.BaseModel, clients_candidate: ValueType
    ) -> ValueType:
        """Validate ``clients`` field."""
        if not clients_candidate:
            raise ValueError("Empty client names prohibited")
        if len(set(clients_candidate)) != len(clients_candidate):
            raise ValueError("Duplicate client names prohibited")

        for name, client_value in clients_candidate.items():
            if client_value.depends_on is not None:
                # check that the "depends_on" target exists.
                if any(
                    [
                        x not in clients_candidate
                        for x in client_value.depends_on.keys()
                    ]
                ):
                    raise ValueError(
                        "Nonexistent dependency in client, " "{0}".format(name)
                    )
                for this_name in [
                    x for x, y in client_value.depends_on.items() if not y
                ]:
                    # default to the last release state from the dependency
                    # client.
                    default_value = clients_candidate[this_name].release_states[
                        -1
                    ]
                    clients_candidate[name].depends_on[
                        this_name
                    ] = default_value  # type: ignore

        return clients_candidate


def load_clients(file_path: pathlib.Path) -> ClientsDefinition:
    """
    Load client definitions from file.

    Args:
        file_path: Path to client definitions file.

    Returns:
        Client definitions.
    Raises:
        ClientsDefinitionError: If an error occurs loading the file.
    """
    result = load_configuration(
        file_path, ClientsDefinition, ClientsDefinitionError, ENTITY_NAME
    )

    return typing.cast(ClientsDefinition, result)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio
import copy
import logging
import typing

import pytest

from foodx_devops_tools.deploy_me._dependency_monitor import (
    DeploymentTerminatedError,
    IterationContext,
    _confirm_dependency_entity_status,
    _generate_dependency_contexts,
    wait_for_dependencies,
)
from foodx_devops_tools.deploy_me._deployment import (
    DeploymentState,
    DeploymentStatus,
    PipelineCliOptions,
    SingularFrameDefinition,
)

MOCK_ITERATION_CONTEXT = IterationContext()
MOCK_ITERATION_CONTEXT.append("some")
MOCK_ITERATION_CONTEXT.append("context")
MOCK_CONTEXT = str(MOCK_ITERATION_CONTEXT)
MOCK_PIPELINE_PARAMETERS = PipelineCliOptions(
    enable_validation=True, monitor_sleep_seconds=0.5, wait_timeout_seconds=0.1
)


class TestGenerateDependencyContexts:
    def test_clean(self):
        mock_dependency_names = {
            "one",
            "two",
        }
        result = _generate_dependency_contexts(
            copy.deepcopy(MOCK_ITERATION_CONTEXT), mock_dependency_names
        )

        one = IterationContext()
        one.append("some")
        one.append("one")
        two = IterationContext()
        two.append("some")
        two.append("two")
        assert result == {
            str(one),
            str(two),
        }


class TestConfirmDependencyFrameStatus:
    @pytest.mark.asyncio
    async def test_clean(self):
        dependency_names = {
            "one",
            "two",
        }
        this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=3)
        await this_status.initialize(MOCK_CONTEXT)
        await this_status.initialize("some.one")
        await this_status.initialize("some.two")

        result = await _confirm_dependency_entity_status(
            dependency_names, this_status, MOCK_ITERATION_CONTEXT, 0.05
        )

        assert result == {
            "some.one",
            "some.two",
        }
        assert (
            await this_status.read("some.context")
        ).code == DeploymentState.ResultType.pending

    @pytest.mark.asyncio
    async def test_delayed_status(self):
        dependency_names = {
            "one",
            "two",
        }
        this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=3)
        await this_status.initialize(MOCK_CONTEXT)
        await this_status.initialize("some.one")

        this_task = asyncio.create_task(
            _confirm_dependency_entity_status(
                dependency_names, this_status, MOCK_ITERATION_CONTEXT, 0.1
            )
        )
        # ensure that the next status change arrives after the first
        # iteration of dependency checks
        await asyncio.sleep(0.2)
        await this_status.initialize("some.two")

        result = await this_task

        assert result == {
            "some.one",
            "some.two",
        }
        assert (
            await this_status.read("some.context")
        ).code == DeploymentState.ResultType.pending

    @pytest.mark.asyncio
    async def test_not_found_raises(self):
        dependency_names = {
            "one",
            "two",
        }
        this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=3)
        this_status.start_monitor()
        await this_status.initialize(MOCK_CONTEXT)
        await this_status.initialize("some.one")

        with pytest.raises(
            DeploymentTerminatedError,
            match=r"^dependency not in entity status",
        ):
            await _confirm_dependency_entity_status(
                dependency_names, this_status, MOCK_ITERATION_CONTEXT, 0.05
            )

        # await this task to allow previous write to propagate
        await asyncio.sleep(0.01)

        assert (
            await this_status.read("some.context")
        ).code == DeploymentState.ResultType.failed


class MockWaiter:
    def __init__(self, sleep_times: typing.List[float]):
        self.sleep_times = sleep_times

    async def __call__(
        self,
        this_context: str,
        frame_data: SingularFrameDefinition,
        frame_status: DeploymentStatus,
        status_monitor_sleep_seconds: int,
    ) -> None:
        for x in self.sleep_times:
            await asyncio.sleep(status_monitor_sleep_seconds)


@pytest.fixture()
def mock_frame_dependency(prep_frame_data):
    async def _apply():
        import foodx_devops_tools.deploy_me._dependency_monitor

        foodx_devops_tools.deploy_me._dependency_monitor.STATUS_KEY_RETRY_SLEEP_SECONDS = (
            0.1
        )
        deployment_data, frame_data = prep_frame_data
        this_context = copy.deepcopy(MOCK_ITERATION_CONTEXT)
        this_context.append("f1")

        frame_data.depends_on = ["df1"]

        this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=2)
        this_status.start_monitor()
        await this_status.initialize(str(this_context))
        for this_df in frame_data.depends_on:
            this_df_context = copy.deepcopy(MOCK_ITERATION_CONTEXT)
            this_df_context.append(this_df)
            await this_status.initialize(str(this_df_context))

        return this_context, this_status, deployment_data, frame_data

    return _apply


class TestProcessDependencies:
    MOCK_SLEEP = 0.1

    @pytest.mark.asyncio
    async def test_dependencies_clean(
        self, caplog, capsys, mock_frame_dependency
    ):
        with caplog.at_level(logging.INFO):
            (
                this_context,
                this_status,
                deployment_data,
                frame_data,
            ) = await mock_frame_dependency()

            # set the dependency status to "success"
            df_context = copy.deepcopy(MOCK_ITERATION_CONTEXT)
            df_context.append("df1")
            await this_status.write(
                str(df_context), DeploymentState.ResultType.success
            )

            await asyncio.wait_for(
                wait_for_dependencies(
                    this_context,
                    frame_data.depends_on,
                    this_status,
                ),
                timeout=1,
            )

        expected_messages = [
            {
                "message": "dependencies completed. proceeding with deployment",
                "log_stdout_both": True,
            },
        ]
        self._check_messages(expected_messages, caplog, capsys)

    async def _change_dependency_state(
        self,
        iteration_context: IterationContext,
        frame_status: DeploymentStatus,
        new_state: DeploymentState.ResultType,
    ):
        await asyncio.sleep(0.5)
        await frame_status.write(str(iteration_context), new_state)

    def _check_messages(self, expected_messages, caplog, capsys):
        captured = capsys.readouterr()
        for this_item in expected_messages:
            this_message = this_item["message"]
            in_both = this_item["log_stdout_both"]
            if this_message not in caplog.text:
                pytest.fail(f"not found in logs, {this_message}")
            if in_both and (this_message not in captured.out):
                pytest.fail(f"not found in stdout, {this_message}")

    @pytest.mark.asyncio
    async def test_in_progress_dependency_succeeds(
        self, caplog, capsys, mock_frame_dependency
    ):
        with caplog.at_level(logging.INFO):
            (
                this_context,
                this_status,
                deployment_data,
                frame_data,
            ) = await mock_frame_dependency()

            # set the dependency status to "success"
            df_context = copy.deepcopy(MOCK_ITERATION_CONTEXT)
            df_context.append("df1")

            asyncio.create_task(
                self._change_dependency_state(
                    df_context, this_status, DeploymentState.ResultType.success
                )
            )

            await asyncio.wait_for(
                wait_for_dependencies(
                    this_context,
                    frame_data.depends_on,
                    this_status,
                ),
                timeout=1,
            )

        expected_messages = [
            {
                "message": "waiting for dependency completion",
                "log_stdout_both": False,
            },
            {
                "message": "dependencies completed. proceeding with deployment",
                "log_stdout_both": True,
            },
        ]
        self._check_messages(expected_messages, caplog, capsys)

    @pytest.mark.asyncio
    async def test_in_progress_dependency_fails(
        self, caplog, capsys, mock_frame_dependency
    ):
        with caplog.at_level(logging.INFO):
            (
                this_context,
                this_status,
                deployment_data,
                frame_data,
            ) = await mock_frame_dependency()

            df_context = copy.deepcopy(MOCK_ITERATION_CONTEXT)
            df_context.append("df1")

            # set the dependency status to "failed"
            asyncio.create_task(
                self._change_dependency_state(
                    df_context, this_status, DeploymentState.ResultType.failed
                )
            )

            with pytest.raises(
                DeploymentTerminatedError,
                match=r"cancelled due to dependency failure",
            ):
                await asyncio.wait_for(
                    wait_for_dependencies(
                        this_context,
                        frame_data.depends_on,
                        this_status,
                    ),
                    timeout=1,
                )

    async def _dependency_late_initialization(
        self,
        iteration_context: IterationContext,
        frame_status: DeploymentStatus,
    ):
        await asyncio.sleep(0.05)
        await frame_status.initialize(str(iteration_context))
        await asyncio.sleep(0.1)
        await frame_status.write(
            str(iteration_context), DeploymentState.ResultType.in_progress
        )
        await asyncio.sleep(0.1)
        await frame_status.write(
            str(iteration_context), DeploymentState.ResultType.success
        )

    @pytest.mark.asyncio
    async def test_missing_dependency_becomes_available(
        self, caplog, capsys, mock_frame_dependency
    ):
        with caplog.at_level(logging.INFO):
            (
                this_context,
                _,
                deployment_data,
                frame_data,
            ) = await mock_frame_dependency()

            df_context = copy.deepcopy(MOCK_ITERATION_CONTEXT)
            df_context.append("df1")

            this_status = DeploymentStatus(MOCK_CONTEXT, timeout_seconds=2)
            this_status.start_monitor()
            await this_status.initialize(str(this_context))

            # set the dependency status to "failed"
            asyncio.create_task(
                self._dependency_late_initialization(df_context, this_status)
            )

            await asyncio.wait_for(
                wait_for_dependencies(
                    this_context,
                    frame_data.depends_on,
                    this_status,
                ),
                timeout=1,
            )

        expected_messages = [
            {
                "message": "dependency not in entity status",
                "log_stdout_both": False,
            },
            {
                "message": "dependencies completed. proceeding with deployment",
                "log_stdout_both": True,
            },
        ]
        self._check_messages(expected_messages, caplog, capsys)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.azure.cloud import (
    AzureCredentials,
    login_service_principal,
)
from foodx_devops_tools.azure.cloud.exceptions import AzureAuthenticationError
from foodx_devops_tools.utilities.exceptions import CommandError

MOCK_CREDENTIALS = AzureCredentials(
    name="mock_user",
    userid="123abc",
    secret="verysecret",
    subscription="this_subscription",
    tenant="this_tenant",
)


@pytest.mark.asyncio
async def test_clean(mock_async_method, mocker):
    mock_return = mocker.MagicMock()
    mock_return.out = """{
  "some": "json output"
}
"""

    mock_run = mock_async_method(
        "foodx_devops_tools.azure.cloud.auth.run_async_command",
        return_value=mock_return,
    )

    result = await login_service_principal(MOCK_CREDENTIALS)

    mock_run.assert_called_once_with(
        [
            "az",
            "login",
            "--service-principal",
            "--username",
            "123abc",
            "--password",
            "verysecret",
            "--tenant",
            "this_tenant",
        ]
    )

    assert result == {
        "some": "json output",
    }


@pytest.mark.asyncio
async def test_commanderror_raises(mocker):
    mock_return = mocker.MagicMock()
    mock_return.out = "some output"

    mocker.patch(
        "foodx_devops_tools.azure.cloud.auth.run_async_command",
        side_effect=CommandError(),
    )

    with pytest.raises(
        AzureAuthenticationError,
        match=r"^Service principal authentication failed",
    ):
        await login_service_principal(MOCK_CREDENTIALS)


@pytest.mark.asyncio
async def test_exception_raises(mocker):
    mock_return = mocker.MagicMock()
    mock_return.out = "some output"

    mocker.patch(
        "foodx_devops_tools.azure.cloud.auth.run_async_command",
        side_effect=RuntimeError(),
    )

    with pytest.raises(
        AzureAuthenticationError,
        match=r"^Service principal authentication failed",
    ):
        await login_service_principal(MOCK_CREDENTIALS)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

from foodx_devops_tools.puff.arm import _linearize_environments


class TestLinearizedEnvironments:
    def test_environments_regions(self):
        mock_base = {
            "this.stub": {
                "p1": "bp1",
                "k2": "bk2",
                "environments": {
                    "e1": {
                        "k1": "e1k1",
                        "regions": [
                            {"r1": {"k1": "e1r1k1"}},
                            {"r2": {"k2": "e1r2k2"}},
                        ],
                    },
                },
            },
        }
        expected_result = {
            "this.stub.e1": {
                "environment": "e1",
                "p1": "bp1",
                "k1": "e1k1",
                "k2": "bk2",
                "regions": [
                    {"r1": {"k1": "e1r1k1"}},
                    {"r2": {"k2": "e1r2k2"}},
                ],
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_multiple_stubs_multiple_environments(self):
        mock_base = {
            "stub1": {
                "p1": "stub1.p1",
                "k2": "stub1.k2",
                # "full" environment
                "environments": {
                    "e1": {
                        "k1": "e1k1",
                    },
                    "e2": {
                        "k1": "e2k1",
                    },
                },
            },
        }
        expected_result = {
            "stub1.e1": {
                "environment": "e1",
                "p1": "stub1.p1",
                "k1": "e1k1",
                "k2": "stub1.k2",
            },
            "stub1.e2": {
                "environment": "e2",
                "p1": "stub1.p1",
                "k1": "e2k1",
                "k2": "stub1.k2",
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_multiple_stubs_mixed_environments(self):
        mock_base = {
            "stub1": {
                "p1": "stub1.p1",
                "k2": "stub1.k2",
                # "full" environment
                "environments": {
                    "e1": {
                        "k1": "e1k1",
                        "regions": [
                            {"r1": {"k1": "e1r1k1"}},
                            {"r2": {"k2": "e1r2k2"}},
                        ],
                    },
                },
            },
            "stub2": {
                "p1": "stub2.p1",
                "k2": "stub2.k2",
                # empty environment
                "environments": dict(),
            },
        }
        expected_result = {
            "stub1.e1": {
                "environment": "e1",
                "p1": "stub1.p1",
                "k1": "e1k1",
                "k2": "stub1.k2",
                "regions": [
                    {"r1": {"k1": "e1r1k1"}},
                    {"r2": {"k2": "e1r2k2"}},
                ],
            },
            "stub2": {
                "p1": "stub2.p1",
                "k2": "stub2.k2",
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_empty_environments(self):
        mock_base = {
            "this.stub": {
                "p1": "bp1",
                "k2": "bk2",
                "environments": dict(),
            },
        }
        expected_result = {
            "this.stub": {
                "p1": "bp1",
                "k2": "bk2",
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_environments_no_environment_regions(self):
        mock_base = {
            "this.stub": {
                "p1": "bp1",
                "k2": "bk2",
                "environments": {
                    "e1": {
                        "k1": "e1k1",
                    },
                },
            },
        }
        expected_result = {
            "this.stub.e1": {
                "environment": "e1",
                "p1": "bp1",
                "k1": "e1k1",
                "k2": "bk2",
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_no_environments_no_regions(self):
        mock_base = {
            "this.stub": {
                "p1": "bp1",
                "k2": "bk2",
            },
        }
        expected_result = {
            "this.stub": {
                "p1": "bp1",
                "k2": "bk2",
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_no_environments_no_regions_empty_base(self):
        mock_base = {
            "this.stub": dict(),
        }
        expected_result = {
            "this.stub": dict(),
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_environments_empty_base(self):
        mock_base = {
            "this.stub": {
                "environments": {
                    "e1": {
                        "k1": "e1k1",
                        "regions": [
                            {"r1": {"k1": "e1r1k1"}},
                            {"r2": {"k2": "e1r2k2"}},
                        ],
                    },
                },
            },
        }
        expected_result = {
            "this.stub.e1": {
                "environment": "e1",
                "k1": "e1k1",
                "regions": [
                    {"r1": {"k1": "e1r1k1"}},
                    {"r2": {"k2": "e1r2k2"}},
                ],
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result

    def test_none_environments(self):
        mock_base = {
            "this.stub": {
                "k1": "k1",
                "environments": {
                    "e1": None,
                },
            },
        }
        expected_result = {
            "this.stub.e1": {
                "environment": "e1",
                "k1": "k1",
            },
        }
        result = _linearize_environments(mock_base)

        assert result == expected_result



#  Copyright (c) 2020 Russell Smiley
#  Copyright (c) 2021 Food-X Technologies
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#
# https://gitlab.com/ci-cd-devops/build_harness/-/blob/main/tests/ci/unit_tests/test_utility.py

import asyncio
from unittest.mock import AsyncMock

import pytest

from foodx_devops_tools.utilities import (
    CapturedStreams,
    run_async_command,
    run_command,
)
from foodx_devops_tools.utilities.exceptions import CommandError


class TestRunCommand:
    def test_simple(self, mocker):
        mock_run = mocker.patch(
            "foodx_devops_tools.utilities.command.subprocess.run"
        )
        command = ["something", "--option"]

        result = run_command(command)

        mock_run.assert_called_once_with(command)
        assert result == mock_run.return_value


class TestRunAsyncCommand:
    @pytest.mark.asyncio
    async def test_simple(self, mocker):
        expected_output = CapturedStreams(out="123", error="no errors")
        async_mock = AsyncMock()
        async_mock.return_value.returncode = 0
        async_mock.return_value.communicate.return_value = (
            expected_output.out.encode(),
            expected_output.error.encode(),
        )
        mock_run = mocker.patch(
            "foodx_devops_tools.utilities.command.asyncio"
            ".create_subprocess_exec",
            side_effect=async_mock,
        )
        command = ["something", "--option"]

        result = await run_async_command(command)

        mock_run.assert_called_once_with(
            *command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE
        )
        assert result == expected_output

    @pytest.mark.asyncio
    async def test_error(self, mocker):
        expected_output = CapturedStreams(out="", error="some error")
        async_mock = AsyncMock()
        async_mock.return_value.returncode = 1
        async_mock.return_value.communicate.return_value = (
            expected_output.out.encode(),
            expected_output.error.encode(),
        )
        mocker.patch(
            "foodx_devops_tools.utilities.command.asyncio"
            ".create_subprocess_exec",
            side_effect=async_mock,
        )
        command = ["something", "--option"]

        with pytest.raises(
            CommandError,
            match=r"^External command run did " r"not exit cleanly",
        ):
            await run_async_command(command)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Azure authentication utilities."""

import asyncio
import dataclasses
import json
import logging
import typing

from foodx_devops_tools.utilities import CapturedStreams, run_async_command
from foodx_devops_tools.utilities.exceptions import CommandError

log = logging.getLogger(__name__)


class AzureAuthenticationError(Exception):
    """Problem authenticating to Azure Cloud."""


@dataclasses.dataclass
class AzureCredentials:
    """Credentials data required for Azure Cloud authentication."""

    name: str
    secret: str
    subscription: str
    tenant: str
    userid: str


def _failure_logging(
    result: typing.Optional[CapturedStreams],
    credentials: AzureCredentials,
    e: Exception,
) -> None:
    log.error(
        "az login failed, {0} ({1}, {2}), {3}".format(
            credentials.name,
            credentials.subscription,
            credentials.tenant,
            str(e),
        )
    )
    if result:
        log.debug(
            "az login stdout, {0} ({1}, {2}), {3}".format(
                credentials.name,
                credentials.subscription,
                credentials.tenant,
                result.out,
            )
        )
        log.debug(
            "az login stderr, {0} ({1}, {2}), {3}".format(
                credentials.name,
                credentials.subscription,
                credentials.tenant,
                result.error,
            )
        )
    else:
        log.error(
            "unexpected error, {0} ({1}, {2}), {3}".format(
                credentials.name,
                credentials.subscription,
                credentials.tenant,
                str(e),
            )
        )


async def login_service_principal(credentials: AzureCredentials) -> dict:
    """
    Login to Azure Cloud using service principal credentials.

    Args:
        credentials: Require Azure Cloud credentials.

    Returns:
        Parsed JSON from ``az login`` output.
    Raises:
        AzureAuthenticationError: If any error occurs during login.
    """
    result = None
    try:
        this_command = [
            "az",
            "login",
            "--service-principal",
            "--username",
            credentials.userid,
            "--password",
            credentials.secret,
            "--tenant",
            credentials.tenant,
        ]

        log.debug(
            f"az login, userid, {credentials.userid}, "
            f"subscription, {credentials.subscription}"
            f"tenant, {credentials.tenant}"
        )
        result = await run_async_command(this_command)

        log.info(
            "login succeeded, {0} ({1}, {2})".format(
                credentials.name, credentials.subscription, credentials.tenant
            )
        )
        result_data = json.loads(result.out)
        return result_data
    except asyncio.CancelledError:
        # should almost always let async cancelled exceptions propagate.
        log.debug("az login cancelled")
        raise
    except CommandError as e:
        _failure_logging(result, credentials, e)
        raise AzureAuthenticationError(
            "Service principal authentication failed, {0}".format(str(e))
        ) from e
    except Exception as e:
        _failure_logging(result, credentials, e)
        raise AzureAuthenticationError(
            "Service principal authentication failed, {0}".format(str(e))
        ) from e



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
import asyncio
import contextlib
import logging
import typing
import uuid

import pytest

from foodx_devops_tools.azure.cloud.resource_group import (
    check_exists as check_resource_group_exists,
)
from foodx_devops_tools.azure.cloud.resource_group import (
    create as create_resource_group,
)
from foodx_devops_tools.azure.cloud.resource_group import (
    delete as delete_resource_group,
)

log = logging.getLogger(__name__)


@pytest.fixture()
def prexisting_resource_group(user_subscription):
    @contextlib.asynccontextmanager
    async def _apply(group_name: str) -> typing.Generator[None, None, None]:
        await create_resource_group(group_name, "West US 2", user_subscription)
        assert await check_resource_group_exists(group_name, user_subscription)

        yield

        # clean up any created resource group on context close
        await asyncio.shield(
            delete_resource_group(group_name, user_subscription)
        )

    return _apply


@pytest.fixture()
def unique_group_name():
    return str(uuid.uuid4())


@pytest.fixture()
def clean_teardown(user_subscription):
    @contextlib.asynccontextmanager
    async def _apply(group_name: str) -> typing.Generator[None, None, None]:
        yield

        await delete_resource_group(group_name, user_subscription)

    return _apply



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Configuration file path management."""

import logging
import pathlib
import typing

from ._exceptions import ConfigurationPathsError

log = logging.getLogger(__name__)

PIPELINE_CONFIG_FILES = {
    "clients.yml",
    "release_states.yml",
    "deployments.yml",
    "frames.yml",
    "puff_map.yml",
    "service_principals.vault",
    "subscriptions.yml",
    "systems.yml",
    "tenants.yml",
}

T = typing.TypeVar("T", bound="PipelineConfigurationPaths")


class PipelineConfigurationPaths:
    """Paths to pipeline configuration files."""

    clients: pathlib.Path
    context: typing.Set[pathlib.Path]
    deployments: pathlib.Path
    frames: pathlib.Path
    puff_map: pathlib.Path
    release_states: pathlib.Path
    service_principals: pathlib.Path
    static_secrets: typing.Set[pathlib.Path]
    subscriptions: pathlib.Path
    systems: pathlib.Path
    tenants: pathlib.Path

    CONFIG_SUBDIRS: typing.Set[str] = {"static_secrets", "context"}

    def __init__(self: T) -> None:
        """Construct ``PipelineConfigurationPaths`` object."""
        for this_file in PIPELINE_CONFIG_FILES:
            path = pathlib.Path(this_file)
            setattr(self, path.stem, path)

    @classmethod
    def from_dict(cls: typing.Type[T], data: dict) -> T:
        """
        Construct ``PipelineConfigurationPaths`` object.

        NOTE: Delivering valid paths is the users responsibility.

        Args:
            data: Dictionary of data to populate in object.
        """
        this_object = cls()
        for x, y in data.items():
            setattr(this_object, x, y)

        return this_object

    @classmethod
    def from_paths(
        cls: typing.Type[T],
        client_config: pathlib.Path,
        system_config: pathlib.Path,
    ) -> T:
        """
        Construct ``PipelineConfigurationPaths`` object.

        Args:
            client_config: Path to client configuration directory.
            system_config: Path to system configuration directory.

        Raises:
            ConfigurationPathsError: If any paths are duplicated between
                client and system.
        """
        this_object = cls()
        client_files = this_object.__acquire_client_files(client_config)
        system_files = this_object.__acquire_system_files(system_config)

        if len(client_files + system_files) > len(PIPELINE_CONFIG_FILES):
            # must be duplicate files between the directories
            log.debug("client files, {0}".format(str(client_files)))
            log.debug("system files, {0}".format(str(system_files)))
            raise ConfigurationPathsError(
                "Duplicate files between "
                "directories, {0}, {1}".format(client_config, system_config)
            )

        this_object.static_secrets = cls.__acquire_static_secrets(client_config)
        this_object.context = cls.__acquire_template_context(
            client_config, system_config
        )

        return this_object

    @staticmethod
    def __acquire_static_secrets(
        client_config: pathlib.Path,
    ) -> typing.Set[pathlib.Path]:
        secrets_path = client_config / "static_secrets"
        result = PipelineConfigurationPaths.__acquire_subdir_files(
            secrets_path, "static secrets"
        )
        return result

    @staticmethod
    def __acquire_template_context(
        client_config: pathlib.Path, system_config: pathlib.Path
    ) -> typing.Set[pathlib.Path]:
        # template context could be located in either client or system config.
        context_client_path = client_config / "context"
        client_context_files = (
            PipelineConfigurationPaths.__acquire_subdir_files(
                context_client_path, "client template context"
            )
        )
        context_system_path = system_config / "context"
        system_context_files = (
            PipelineConfigurationPaths.__acquire_subdir_files(
                context_system_path, "system template context"
            )
        )
        result = client_context_files.union(system_context_files)
        return result

    def __acquire_client_files(
        self: T, client_config: pathlib.Path
    ) -> typing.List[pathlib.Path]:
        client_files = list()
        for x in client_config.iterdir():
            if (
                x.is_file()
                and (x.name in PIPELINE_CONFIG_FILES)
                and (x.stem not in self.CONFIG_SUBDIRS)
            ):
                log.info("adding client configuration file, {0}".format(x))
                setattr(self, x.stem, x)
                client_files.append(x)

        return client_files

    def __acquire_system_files(
        self: T, system_config: pathlib.Path
    ) -> typing.List[pathlib.Path]:
        system_files = list()
        for x in system_config.iterdir():
            if x.is_file() and (x.name in PIPELINE_CONFIG_FILES):
                log.info("adding system configuration file, {0}".format(x))
                setattr(self, x.stem, x)
                system_files.append(x)

        return system_files

    @staticmethod
    def __acquire_subdir_files(
        this_path: pathlib.Path, category: str
    ) -> typing.Set[pathlib.Path]:
        result = set()
        if this_path.is_dir():
            log.debug(
                "{2} directory, {0}, {1}".format(
                    this_path, str(list(this_path.iterdir())), category
                )
            )
            for x in this_path.iterdir():
                if x.is_file():
                    log.info(
                        "adding file to configuration, {1}, {0}".format(
                            x, category
                        )
                    )
                    result.add(x)
        elif this_path.exists():
            log.debug(f"{category} not a directory, {this_path}")
        else:
            log.debug(f"{category} not present, {this_path}")

        return result



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.deploy_me._deployment import do_deploy
from foodx_devops_tools.pipeline_config import PipelineConfiguration
from tests.ci.support.pipeline_config import MOCK_RESULTS


@pytest.fixture()
def prep_data(mock_async_method, mock_flattened_deployment, mock_run_puff):
    pipeline_config = PipelineConfiguration.parse_obj(MOCK_RESULTS)

    deployment_data = mock_flattened_deployment[0]
    deployment_data.data.iteration_context.append("f1")
    deployment_data.context.frame_name = "f1"
    deployment_data.context.release_state = "r1"
    deployment_data.context.azure_subscription_name = "sub1"

    mock_frame = mock_async_method(
        "foodx_devops_tools.deploy_me._deployment.deploy_frame"
    )

    return mock_frame, deployment_data, pipeline_config


class TestDoDeploy:
    @pytest.mark.asyncio
    async def test_validation_clean(
        self,
        mocker,
        prep_data,
        mock_completion_event,
        mock_login,
        pipeline_parameters,
    ):
        cli_options = pipeline_parameters()
        mock_frame, deployment_data, pipeline_config = prep_data

        await do_deploy(
            pipeline_config,
            deployment_data,
            cli_options,
        )

        mock_frame.assert_called_once_with(
            pipeline_config.frames.frames["f1"],
            mocker.ANY,
            mocker.ANY,
            cli_options,
        )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Git related utilities."""

from .command import run_command


def get_sha() -> str:
    """Get the git commit SHA of HEAD."""
    result = run_command(
        ["git", "rev-parse", "HEAD"], text=True, capture_output=True
    )

    if result.returncode != 0:
        raise RuntimeError("Git commit SHA acquisition failed")
    else:
        this_sha = result.stdout[0:10]

    return this_sha



#  Copyright (c) 2022 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Utilities for maintaining timing information."""

import contextlib
import datetime
import logging
import time
import typing

log = logging.getLogger(__name__)

T = typing.TypeVar("T", bound="Timer")


class Timer:
    """A simple relative monotonic timer."""

    def __init__(self: T) -> None:
        """Construct ``Timer`` object."""
        self.start_time_seconds: float = 0.0
        self.stop_time_seconds: float = 0.0

    def start(self: T) -> None:
        """Record the start time."""
        self.start_time_seconds = time.monotonic()

    def stop(self: T) -> None:
        """Record the stop time."""
        self.stop_time_seconds = time.monotonic()

    @property
    def elapsed_time_seconds(self: T) -> float:
        """Calculate the elapsed time interval in seconds."""
        return self.stop_time_seconds - self.start_time_seconds

    @property
    def elapsed_time_formatted(self: T) -> str:
        """Calculate the elapsed time and format into a string."""
        return str(datetime.timedelta(seconds=self.elapsed_time_seconds))

    def log_duration(
        self: T, this_log: logging.Logger, iteration_context: str
    ) -> None:
        """Log the elapsed time of the iteration."""
        this_log.info(
            f"{iteration_context} "
            f"elapsed time, {self.elapsed_time_formatted}"
        )


@contextlib.contextmanager
def timing(
    this_log: logging.Logger,
    iteration_context: str,
) -> typing.Generator[Timer, None, Timer]:
    """Manage the context of calculating a time interval."""
    this_timer = Timer()
    this_timer.start()

    yield this_timer

    this_timer.stop()
    this_timer.log_duration(this_log, iteration_context)

    return this_timer



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Jinja2 template application."""

import pathlib
import typing

import aiofiles
import jinja2

TemplateParameters = typing.Dict[str, typing.Any]

T = typing.TypeVar("T", bound="FrameTemplates")


class FrameTemplates:
    """Manage the template environment for a frame."""

    environment: jinja2.Environment

    def __init__(
        self: T, template_search_paths: typing.List[pathlib.Path]
    ) -> None:
        """
        Construct ``FrameTemplates`` object.

        Args:
            template_search_paths: Directory paths where templates may be found.
        """
        self.environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(template_search_paths),
            autoescape=jinja2.select_autoescape(),
        )

    async def apply_template(
        self: T,
        source_template: str,
        target_path: pathlib.Path,
        parameters: TemplateParameters,
    ) -> None:
        """
        Apply jinja2 template to a target file.

        Args:
            source_template: Name of jinja2 template.
            target_path: Target fulfilled file.
            parameters: Parameters to be consumed by the template.
        """
        template = self.environment.get_template(source_template)
        content = template.render(**parameters)
        async with aiofiles.open(target_path, mode="w") as f:
            await f.write(content)


def apply_dynamic_template(
    source_template: str,
    parameters: TemplateParameters,
) -> str:
    """
    Apply parameter data to jinja2 template string.

    Args:
        source_template: Template text to be processed.
        parameters: Parameter context data to be made available to the template.

    Returns:
        Text result of template processing.
    """
    template = jinja2.Template(source_template)
    content = template.render(**parameters)

    return content



#  Copyright (c) 2022 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import logging

import click

from foodx_devops_tools.pipeline_config import FlattenedDeployment
from foodx_devops_tools.pipeline_config.frames import ApplicationStepScript
from foodx_devops_tools.utilities.command import run_async_command
from foodx_devops_tools.utilities.jinja2 import apply_dynamic_template

log = logging.getLogger(__name__)


async def script_step(
    this_step: ApplicationStepScript,
    deployment_data: FlattenedDeployment,
) -> None:
    """
    Run command line commands in Bash shell.

    Args:
        this_step: Deployment definition for this step action.
        deployment_data: Deployment context related parameters.
    """
    this_context = str(deployment_data.data.iteration_context)
    step_context = f"{this_context}.{this_step.name}"

    log.debug(
        f"deployment_data.context, "
        f"{step_context}, {str(deployment_data.context)}"
    )
    log.debug(
        f"deployment_data.data, {step_context}, "
        f"{str(deployment_data.data.deployment_tuple)}, "
        f"{str(deployment_data.data.location_primary)}, "
        f"{str(deployment_data.data.location_secondary)}, "
        f"{str(deployment_data.data.release_state)}, "
        f"{str(deployment_data.data.iteration_context)}"
    )
    try:
        template_parameters = deployment_data.construct_template_parameters()
        log.debug(f"template parameters, {step_context}, {template_parameters}")

        templated_inline_script = apply_dynamic_template(
            this_step.script, template_parameters
        )
        this_command = [
            "/bin/bash",
            "-c",
            templated_inline_script,
        ]
        await run_async_command(this_command)

    except Exception as e:
        message = f"step shell script failed, {step_context}, {str(e)}"
        log.error(message)
        click.echo(click.style(message, fg="red"), err=True)
        raise



#  Copyright (c) 2022 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import logging
import time

from foodx_devops_tools.profiling import timing

log = logging.getLogger(__name__)


def test_clean(mocker):
    mocker.patch(
        "foodx_devops_tools.profiling.time.monotonic", side_effect=[1.2, 3.5]
    )

    with timing(log, "some.context") as t:
        time.sleep(0.2)

    assert t.elapsed_time_seconds == 2.3
    assert t.elapsed_time_formatted == "0:00:02.300000"



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import logging
import os

import pytest

from foodx_devops_tools.pipeline_config import (
    ClientsDefinition,
    DeploymentsDefinition,
    PipelineConfiguration,
    PipelineConfigurationPaths,
    ServicePrincipals,
    StaticSecrets,
    SubscriptionsDefinition,
)
from foodx_devops_tools.pipeline_config.exceptions import (
    PipelineConfigurationError,
)
from tests.ci.support.pipeline_config import (
    CLEAN_SPLIT,
    MOCK_PATHS,
    MOCK_RESULTS,
    MOCK_SECRET,
    split_directories,
)

log = logging.getLogger(__name__)


class TestPipelineConfiguration:
    def test_default(self, mock_loads, mock_results):
        mock_loads(mock_results)
        under_test = PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

        assert len(under_test.clients) == 2
        assert "c1" in under_test.clients
        assert len(under_test.release_states) == 2
        assert under_test.release_states[0] == "r1"
        assert len(under_test.subscriptions) == 1
        assert (
            under_test.subscriptions["sys1_c1_r1a"].ado_service_connection
            == "some-name"
        )
        assert under_test.subscriptions["sys1_c1_r1a"].azure_id == "abc123"
        assert len(under_test.systems) == 2
        assert under_test.systems[0] == "sys1"

        assert len(under_test.tenants) == 1
        assert under_test.tenants["t1"].azure_id == "123abc"

    def test_bad_deployment_name_raises(self, mock_loads, mock_results):
        mock_results.deployments = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "deployment_tuples": {
                        "badname": {
                            "subscriptions": {
                                "sys1_c1_r1a": {
                                    "locations": [
                                        {"primary": "l1"},
                                        {"primary": "l2"},
                                    ],
                                    "root_fqdn": "some.where",
                                },
                            },
                        },
                    },
                    "url_endpoints": ["a", "p"],
                },
            }
        ).deployments
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad deployment tuple"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

    def test_bad_deployment_clients_raises(self, mock_loads, mock_results):
        mock_results.deployments = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "deployment_tuples": {
                        "sys1-bad-r1": {
                            "subscriptions": {
                                "sys1_c1_r1a": {
                                    "locations": [
                                        {"primary": "l1"},
                                        {"primary": "l2"},
                                    ],
                                    "root_fqdn": "some.where",
                                }
                            },
                        },
                    },
                    "url_endpoints": ["a", "p"],
                }
            }
        ).deployments
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad client in deployment tuple"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

    def test_bad_deployment_release_states_raises(
        self, mock_loads, mock_results
    ):
        mock_results.deployments = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "deployment_tuples": {
                        "s1-c1-bad": {
                            "subscriptions": {
                                "subname": {
                                    "locations": [
                                        {"primary": "l1"},
                                        {"primary": "l2"},
                                    ],
                                    "root_fqdn": "some.where",
                                }
                            },
                        },
                    },
                    "url_endpoints": ["a", "p"],
                }
            }
        ).deployments
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError,
            match=r"Bad release state in deployment tuple",
        ):
            PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

    def test_bad_deployment_systems_raises(self, mock_loads, mock_results):
        mock_results.deployments = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "deployment_tuples": {
                        "bad-c1-r1": {
                            "subscriptions": {
                                "subname": {
                                    "locations": [
                                        {"primary": "l1"},
                                        {"primary": "l2"},
                                    ],
                                    "root_fqdn": "some.where",
                                }
                            },
                        },
                    },
                    "url_endpoints": ["a", "p"],
                }
            }
        ).deployments
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad system in deployment tuple"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

    def test_bad_client_release_states_raises(self, mock_loads, mock_results):
        mock_results.clients = ClientsDefinition.parse_obj(
            {
                "clients": {
                    "c1": {"release_states": ["bad_state"], "system": "s1"},
                },
            }
        ).clients
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad release state in client"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

    def test_bad_client_system_raises(self, mock_loads, mock_results):
        mock_results.clients = ClientsDefinition.parse_obj(
            {
                "clients": {
                    "c1": {"release_states": ["r1"], "system": "bad_system"},
                },
            }
        ).clients
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad system in client"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

    def test_bad_subscription_tenant_raises(self, mock_loads, mock_results):
        mock_results.subscriptions = SubscriptionsDefinition.parse_obj(
            {
                "subscriptions": {
                    "sys1_c1_r1a": {
                        "ado_service_connection": "some-name",
                        "azure_id": "abc123",
                        "tenant": "bad_name",
                    },
                },
            }
        ).subscriptions
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad tenant\(s\) in subscription"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

    def test_load_files(self):
        with split_directories(CLEAN_SPLIT.copy()) as (
            client_path,
            system_path,
        ):
            client_config = client_path / "configuration"
            system_config = system_path / "configuration"
            this_paths = PipelineConfigurationPaths.from_paths(
                client_config, system_config
            )
            # just expect no exceptions, for now.
            PipelineConfiguration.from_files(this_paths, MOCK_SECRET)

    def test_none_token_clean(self):
        with split_directories(CLEAN_SPLIT.copy()) as (
            client_path,
            system_path,
        ):
            client_config = client_path / "configuration"
            system_config = system_path / "configuration"
            this_paths = PipelineConfigurationPaths.from_paths(
                client_config, system_config
            )
            # just expect no exceptions, for now.
            PipelineConfiguration.from_files(this_paths, None)

    def test_none_token_raises(self):
        """Missing service principals file raises."""
        with split_directories(CLEAN_SPLIT.copy()) as (
            client_path,
            system_path,
        ):
            client_config = client_path / "configuration"
            system_config = system_path / "configuration"
            this_paths = PipelineConfigurationPaths.from_paths(
                client_config, system_config
            )
            # ensure the vault file doesn't exist
            os.remove(client_config / "service_principals.vault")
            with pytest.raises(
                FileNotFoundError,
                match=r"^Missing service principals vault file",
            ):
                PipelineConfiguration.from_files(this_paths, None)

    def test_load_dict(self):
        PipelineConfiguration.parse_obj(MOCK_RESULTS.copy())


class TestDeploymentSubscriptions:
    def test_deployment_subscription_not_in_subscriptions_raises(
        self, mock_loads, mock_results
    ):
        """Deployment declares a subscription not present in subscriptions."""
        mock_results.deployments = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "deployment_tuples": {
                        "sys1-c1-r1": {
                            "subscriptions": {
                                "bad_name": {
                                    "locations": [
                                        {"primary": "l1"},
                                        {"primary": "l2"},
                                    ],
                                    "root_fqdn": "some.where",
                                }
                            },
                        },
                    },
                    "url_endpoints": ["a", "p"],
                }
            }
        ).deployments
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError,
            match=r"Deployment subscription not defined in subscriptions",
        ):
            PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

    def test_deployment_subscription_not_in_service_principals_raises(
        self, mock_loads, mock_results
    ):
        """Deployment declares a subscription not present in service
        principals."""
        mock_results.service_principals = ServicePrincipals.parse_obj(
            {
                "service_principals": {
                    "inactive_subscription": {
                        "id": "12345",
                        "secret": "verysecret",
                        "name": "sp_name",
                    },
                },
            }
        ).service_principals
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError,
            match=r"Deployment subscription not defined in service principals",
        ):
            PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)

    def test_deployment_subscription_not_in_static_secrets_raises(
        self, mock_loads, mock_results
    ):
        """Deployment declares a subscription not present in static secrets."""
        mock_results.static_secrets = StaticSecrets.parse_obj(
            {
                "static_secrets": {
                    "inactive_subscription": {"k1": "k1v"},
                },
            }
        ).static_secrets
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError,
            match=r"Deployment subscription not defined in static secrets",
        ):
            PipelineConfiguration.from_files(MOCK_PATHS, MOCK_SECRET)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import re

from ..patterns import BRANCH_PREFIX, SEMANTIC_VERSION_GITREF, TAG_PREFIX
from ..release import ReleaseState
from ._exceptions import ReleaseStateError

SEMANTIC_TAG_GITREF = r"{0}{1}".format(TAG_PREFIX, SEMANTIC_VERSION_GITREF)

REGEX_TABLE = {
    ReleaseState.ftr.name: r"^({0}feature/)|(refs/pull/)".format(BRANCH_PREFIX),
    ReleaseState.dev.name: r"^{0}main$".format(BRANCH_PREFIX),
    ReleaseState.qa.name: r"^{0}-alpha\.\d+$".format(SEMANTIC_TAG_GITREF),
    ReleaseState.stg.name: r"^{0}-beta\.\d+$".format(SEMANTIC_TAG_GITREF),
    ReleaseState.prd.name: r"^{0}$".format(SEMANTIC_TAG_GITREF),
}


def identify_release_state(git_ref: str) -> ReleaseState:
    """
    Parse git reference to map to release states.

    Args:
        git_ref: git reference to parse

    Returns:
        Parsed release state.
    Raises:
        ReleaseStateError: If no release state can be inferred.
    """
    found_state = False
    state_found = None
    for label, pattern in REGEX_TABLE.items():
        if re.match(pattern, git_ref) is not None:
            found_state = True
            state_found = ReleaseState[label]

    if (not found_state) or (not state_found):
        raise ReleaseStateError(
            "Unable to match git reference to any release state, {0}".format(
                git_ref
            )
        )

    return state_found



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy
import pathlib

import pytest

from foodx_devops_tools.pipeline_config.views import FlattenedDeployment
from foodx_devops_tools.utilities.templates import TemplateFiles, TemplatePaths


@pytest.fixture()
def mock_test_data(mock_flattened_deployment):
    under_test = copy.deepcopy(mock_flattened_deployment[0])
    under_test.context.application_name = "this_app"
    under_test.context.frame_name = "f1"
    under_test.data.frame_folder = pathlib.Path("frame/folder")

    return copy.deepcopy(under_test)


class TestConstructDeploymentPaths:
    MOCK_CONTEXT = "some_context"

    def _do_check(self, parameters: dict, under_test: FlattenedDeployment):
        result = under_test.construct_deployment_paths(*parameters["input"])

        assert result == parameters["expected"]

    def test_none_arm_file_puff_file(self, mock_test_data):
        parameters = {
            "input": [
                # specified_arm_file
                None,
                # specified_puff_file
                None,
            ],
            "expected": TemplateFiles(
                arm_template=TemplatePaths(
                    source=pathlib.Path("frame/folder/this_app.json"),
                    target=pathlib.Path("frame/folder/working/w/this_app.json"),
                ),
                arm_template_parameters=TemplatePaths(
                    source=pathlib.Path("frame/folder/this_app.yml"),
                    target=pathlib.Path(
                        "frame/folder/working/w/this_app.c1.sys1_c1_r1a.json"
                    ),
                ),
            ),
        }

        self._do_check(parameters, mock_test_data)

    def test_arm_file_none_puff_file(self, mock_test_data):
        parameters = {
            "input": [
                # specified_arm_file
                pathlib.Path("arm_file.json"),
                # specified_puff_file
                None,
            ],
            "expected": TemplateFiles(
                arm_template=TemplatePaths(
                    source=pathlib.Path("frame/folder/arm_file.json"),
                    target=pathlib.Path("frame/folder/working/w/arm_file.json"),
                ),
                arm_template_parameters=TemplatePaths(
                    source=pathlib.Path("frame/folder/arm_file.yml"),
                    target=pathlib.Path(
                        "frame/folder/working/w/arm_file.c1.sys1_c1_r1a.json"
                    ),
                ),
            ),
        }

        self._do_check(parameters, mock_test_data)

    def test_pathed_arm_file_none_puff_file(self, mock_test_data):
        parameters = {
            "input": [
                # specified_arm_file
                pathlib.Path("sub/arm_file.json"),
                # specified_puff_file
                None,
            ],
            "expected": TemplateFiles(
                arm_template=TemplatePaths(
                    source=pathlib.Path("frame/folder/sub/arm_file.json"),
                    target=pathlib.Path("frame/folder/working/w/arm_file.json"),
                ),
                arm_template_parameters=TemplatePaths(
                    source=pathlib.Path("frame/folder/sub/arm_file.yml"),
                    target=pathlib.Path(
                        "frame/folder/working/w/arm_file.c1.sys1_c1_r1a.json"
                    ),
                ),
            ),
        }

        self._do_check(parameters, mock_test_data)

    def test_pathed_arm_parameters_file(self, mock_test_data):
        parameters = {
            "input": [
                # specified_arm_file
                None,
                # specified_puff_file
                None,
            ],
            "expected": TemplateFiles(
                arm_template=TemplatePaths(
                    source=pathlib.Path("frame/folder/this_app.json"),
                    target=pathlib.Path("frame/folder/working/w/this_app.json"),
                ),
                arm_template_parameters=TemplatePaths(
                    source=pathlib.Path("frame/folder/this_app.yml"),
                    target=pathlib.Path(
                        "frame/folder/working/w/this_app.c1.sys1_c1_r1a.json"
                    ),
                ),
            ),
        }

        self._do_check(parameters, mock_test_data)

    def test_independent_arm_file_puff_file(self, mock_test_data):
        parameters = {
            "input": [
                # specified_arm_file
                pathlib.Path("arm_file.json"),
                # specified_puff_file
                pathlib.Path("puff_file.yml"),
            ],
            "expected": TemplateFiles(
                arm_template=TemplatePaths(
                    source=pathlib.Path("frame/folder/arm_file.json"),
                    target=pathlib.Path("frame/folder/working/w/arm_file.json"),
                ),
                arm_template_parameters=TemplatePaths(
                    source=pathlib.Path("frame/folder/puff_file.yml"),
                    target=pathlib.Path(
                        "frame/folder/working/w/puff_file.c1.sys1_c1_r1a.json"
                    ),
                ),
            ),
        }

        self._do_check(parameters, mock_test_data)

    def test_pathed_independent_arm_file_puff_file(self, mock_test_data):
        parameters = {
            "input": [
                # specified_arm_file
                pathlib.Path("arm_file.json"),
                # specified_puff_file
                pathlib.Path("puff/puff_file.yml"),
            ],
            "expected": TemplateFiles(
                arm_template=TemplatePaths(
                    source=pathlib.Path("frame/folder/arm_file.json"),
                    target=pathlib.Path("frame/folder/working/w/arm_file.json"),
                ),
                arm_template_parameters=TemplatePaths(
                    source=pathlib.Path("frame/folder/puff/puff_file.yml"),
                    target=pathlib.Path(
                        "frame/folder/working/w/puff_file.c1.sys1_c1_r1a.json"
                    ),
                ),
            ),
        }

        self._do_check(parameters, mock_test_data)

    def test_out_of_frame_nonjinja_arm_file(self, mock_test_data):
        parameters = {
            "input": [
                # specified_arm_file
                pathlib.Path("../common/files/arm_file.json"),
                # specified_puff_file
                pathlib.Path("puff_file.yml"),
            ],
            "expected": TemplateFiles(
                arm_template=TemplatePaths(
                    source=pathlib.Path(
                        "frame/folder/../common/files/arm_file.json"
                    ),
                    target=pathlib.Path("frame/folder/working/w/arm_file.json"),
                ),
                arm_template_parameters=TemplatePaths(
                    source=pathlib.Path("frame/folder/puff_file.yml"),
                    target=pathlib.Path(
                        "frame/folder/working/w/puff_file.c1.sys1_c1_r1a.json"
                    ),
                ),
            ),
        }

        self._do_check(parameters, mock_test_data)

    def test_out_of_frame_jinja_arm_file(self, mock_test_data):
        parameters = {
            "input": [
                # specified_arm_file
                pathlib.Path("../common/files/jinja2.arm_file.json"),
                # specified_puff_file
                pathlib.Path("puff_file.yml"),
            ],
            "expected": TemplateFiles(
                arm_template=TemplatePaths(
                    source=pathlib.Path(
                        "frame/folder/../common/files/jinja2.arm_file.json"
                    ),
                    target=pathlib.Path(
                        "frame/folder/working/w/jinja2.arm_file.json"
                    ),
                ),
                arm_template_parameters=TemplatePaths(
                    source=pathlib.Path("frame/folder/puff_file.yml"),
                    target=pathlib.Path(
                        "frame/folder/working/w/puff_file.c1.sys1_c1_r1a.json"
                    ),
                ),
            ),
        }

        self._do_check(parameters, mock_test_data)

    def test_out_of_frame_nonjinja_puff_file(self, mock_test_data):
        parameters = {
            "input": [
                # specified_arm_file
                None,
                # specified_puff_file
                pathlib.Path("../common/files/puff_file.yml"),
            ],
            "expected": TemplateFiles(
                arm_template=TemplatePaths(
                    source=pathlib.Path("frame/folder/this_app.json"),
                    target=pathlib.Path("frame/folder/working/w/this_app.json"),
                ),
                arm_template_parameters=TemplatePaths(
                    source=pathlib.Path(
                        "frame/folder/../common/files/puff_file.yml"
                    ),
                    target=pathlib.Path(
                        "frame/folder/working/w/puff_file.c1.sys1_c1_r1a.json"
                    ),
                ),
            ),
        }

        self._do_check(parameters, mock_test_data)

    def test_out_of_frame_jinja_puff_file(self, mock_test_data):
        parameters = {
            "input": [
                # specified_arm_file
                None,
                # specified_puff_file
                pathlib.Path("../common/files/jinja2.puff_file.yml"),
            ],
            "expected": TemplateFiles(
                arm_template=TemplatePaths(
                    source=pathlib.Path("frame/folder/this_app.json"),
                    target=pathlib.Path("frame/folder/working/w/this_app.json"),
                ),
                arm_template_parameters=TemplatePaths(
                    source=pathlib.Path(
                        "frame/folder/../common/files/jinja2.puff_file.yml"
                    ),
                    target=pathlib.Path(
                        "frame/folder/working/w/jinja2.puff_file.c1.sys1_c1_r1a"
                        ".json"
                    ),
                ),
            ),
        }

        self._do_check(parameters, mock_test_data)

    def test_pathed_jinja_puff_file(self, mock_test_data):
        parameters = {
            "input": [
                # specified_arm_file
                None,
                # specified_puff_file
                pathlib.Path("sub/jinja2.puff_file.yml"),
            ],
            "expected": TemplateFiles(
                arm_template=TemplatePaths(
                    source=pathlib.Path("frame/folder/this_app.json"),
                    target=pathlib.Path("frame/folder/working/w/this_app.json"),
                ),
                arm_template_parameters=TemplatePaths(
                    source=pathlib.Path(
                        "frame/folder/sub/jinja2.puff_file.yml"
                    ),
                    target=pathlib.Path(
                        "frame/folder/working/w/jinja2.puff_file.c1.sys1_c1_r1a"
                        ".json"
                    ),
                ),
            ),
        }

        self._do_check(parameters, mock_test_data)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Manage pipeline configuration files collectively."""

import logging
import re
import typing

import pydantic

from ._exceptions import PipelineConfigurationError
from ._paths import PipelineConfigurationPaths
from .clients import ValueType as ClientsData
from .clients import load_clients
from .deployments import DeploymentsEndpointsDefinitions as DeploymentsData
from .deployments import load_deployments
from .frames import ValueType as FramesData
from .frames import load_frames
from .release_states import ValueType as ReleaseStatesData
from .release_states import load_release_states
from .service_principals import ServicePrincipals
from .service_principals import ValueType as ServicePrincipalsData
from .service_principals import load_service_principals
from .static_secrets import StaticSecrets
from .static_secrets import ValueType as StaticSecretsData
from .static_secrets import load_static_secrets
from .subscriptions import ValueType as SubscriptionsData
from .subscriptions import load_subscriptions
from .systems import ValueType as SystemsData
from .systems import load_systems
from .template_context import ValueType as TemplateContextData
from .template_context import load_template_context
from .tenants import ValueType as TenantsData
from .tenants import load_tenants

log = logging.getLogger(__name__)

DEPLOYMENT_NAME_REGEX = (
    r"^(?P<system>[a-z0-9]+)"
    r"[_-]"
    r"(?P<client>[a-z0-9]+)"
    r"[_-]"
    r"(?P<release_state>[a-z0-9]+)$"
)


T = typing.TypeVar("T", bound="PipelineConfiguration")


class PipelineConfiguration(pydantic.BaseModel):
    """Pipeline configuration data."""

    clients: ClientsData
    context: TemplateContextData
    release_states: ReleaseStatesData
    deployments: DeploymentsData
    frames: FramesData
    service_principals: typing.Optional[ServicePrincipalsData]
    static_secrets: typing.Optional[StaticSecretsData]
    subscriptions: SubscriptionsData
    systems: SystemsData
    tenants: TenantsData

    decrypt_token: typing.Optional[str] = None

    @staticmethod
    def __load_encrypted_data(
        paths: PipelineConfigurationPaths, decrypt_token: typing.Optional[str]
    ) -> typing.Tuple[
        typing.Optional[ServicePrincipals], typing.Optional[StaticSecrets]
    ]:
        service_principals_config = None
        static_secrets_config = None
        if decrypt_token:
            service_principals_config = load_service_principals(
                paths.service_principals, decrypt_token
            )
            static_secrets_config = load_static_secrets(
                paths.static_secrets, decrypt_token
            )
        else:
            if not paths.service_principals.is_file():
                raise FileNotFoundError(
                    "Missing service principals vault "
                    "file, {0}".format(paths.service_principals)
                )
            missing_files = {
                str(x) for x in paths.static_secrets if not x.is_file()
            }
            if any(missing_files):
                raise FileNotFoundError(
                    "Missing static secrets files, {0}".format(
                        str(missing_files)
                    )
                )

        return service_principals_config, static_secrets_config

    @classmethod
    def from_files(
        cls: typing.Type[T],
        paths: PipelineConfigurationPaths,
        decrypt_token: typing.Optional[str],
    ) -> T:
        """
        Load pipeline configuration from collection of files.

        Args:
            paths: Paths to pipeline configuration files.

        Returns:
            Instantiated ``PipelineConfiguration`` object.
        """
        client_config = load_clients(paths.clients)
        deployment_config = load_deployments(paths.deployments)
        frames_config = load_frames(paths.frames)
        release_state_config = load_release_states(paths.release_states)
        subscription_config = load_subscriptions(paths.subscriptions)
        system_config = load_systems(paths.systems)
        tenant_config = load_tenants(paths.tenants)
        template_context = load_template_context(paths.context)

        (
            service_principals_config,
            static_secrets_config,
        ) = cls.__load_encrypted_data(paths, decrypt_token)

        kwargs = {
            "clients": client_config.clients,
            "release_states": release_state_config.release_states,
            "decrypt_token": decrypt_token,
            "deployments": deployment_config.deployments,
            "frames": frames_config.frames,
            "subscriptions": subscription_config.subscriptions,
            "systems": system_config.systems,
            "tenants": tenant_config.tenants,
            "context": template_context.context,
        }
        if service_principals_config:
            kwargs[
                "service_principals"
            ] = service_principals_config.service_principals
        if static_secrets_config:
            kwargs["static_secrets"] = static_secrets_config.static_secrets

        new_object = cls(**kwargs)

        return new_object

    @pydantic.root_validator()
    def check_clients(cls: pydantic.BaseModel, loaded_data: dict) -> dict:
        """Cross-check clients data with other fields."""
        for name, data in loaded_data["clients"].items():
            if not all(
                [
                    x in loaded_data["release_states"]
                    for x in data.release_states
                ]
            ):
                message = "Bad release state in client, {0}".format(name)
                log.error(
                    "{0}, {1}, {2}".format(
                        message,
                        str(loaded_data["release_states"]),
                        str(data.release_states),
                    )
                )
                raise PipelineConfigurationError(message)

            if data.system not in loaded_data["systems"]:
                message = "Bad system in client, {0}".format(name)
                log.error(
                    "{0}, {1}, {2}".format(
                        message,
                        str(loaded_data["release_states"]),
                        str(data.release_states),
                    )
                )
                raise PipelineConfigurationError(message)
        return loaded_data

    @staticmethod
    def __check_deployment_tuple(
        deployment_tuple: str, loaded_data: dict
    ) -> None:
        """Check for a valid deployment tuple in deployments."""
        this_re = re.compile(DEPLOYMENT_NAME_REGEX)
        result = this_re.match(deployment_tuple)

        if not result:
            message = "Bad deployment tuple, {0}".format(deployment_tuple)
            log.error("{0}, {1}".format(message, DEPLOYMENT_NAME_REGEX))
            raise PipelineConfigurationError(message)

        this_client = result.group("client")
        if this_client not in loaded_data["clients"]:
            message = "Bad client in deployment tuple, {0}".format(this_client)
            log.error(
                "{0}, {1}, {2}".format(
                    message, deployment_tuple, str(loaded_data["clients"])
                )
            )
            raise PipelineConfigurationError(message)

        this_release_state = result.group("release_state")
        if this_release_state not in loaded_data["release_states"]:
            message = "Bad release state in deployment tuple, {0}".format(
                this_release_state
            )
            log.error(
                "{0}, {1}, {2}".format(
                    message,
                    deployment_tuple,
                    str(loaded_data["release_states"]),
                )
            )
            raise PipelineConfigurationError(message)

        this_system = result.group("system")
        if this_system not in loaded_data["systems"]:
            message = "Bad system in deployment tuple, {0}".format(this_system)
            log.error(
                "{0}, {1}, {2}".format(
                    message, deployment_tuple, str(loaded_data["systems"])
                )
            )
            raise PipelineConfigurationError(message)

    @staticmethod
    def __report_deployment_subscription_error(
        deployment_tuple: str,
        subscription_name: str,
        subscriptions: set,
        category: str,
    ) -> None:
        message = (
            "Deployment subscription not defined in "
            "{0}, {1}, {2}".format(
                category, deployment_tuple, subscription_name
            )
        )
        log.error(
            "{0}, {1}".format(
                message,
                str(subscriptions),
            )
        )
        raise PipelineConfigurationError(message)

    @staticmethod
    def __check_deployment_subscriptions(
        deployment_tuple: str, loaded_data: dict
    ) -> None:
        """Check for valid subscriptions in deployments."""
        deployment_subscriptions = set(
            loaded_data["deployments"]
            .deployment_tuples[deployment_tuple]
            .subscriptions.keys()
        )
        for this_deployment_subscription in deployment_subscriptions:
            if this_deployment_subscription not in loaded_data["subscriptions"]:
                # a deployment subscription must be defined in subscriptions.yml
                PipelineConfiguration.__report_deployment_subscription_error(
                    deployment_tuple,
                    this_deployment_subscription,
                    set(loaded_data["subscriptions"].keys()),
                    "subscriptions",
                )

            if (
                ("service_principals" in loaded_data)
                and (loaded_data["service_principals"])
                and (
                    this_deployment_subscription
                    not in loaded_data["service_principals"]
                )
            ):
                # a deployment subscription must be defined in
                # service_principals.vault
                PipelineConfiguration.__report_deployment_subscription_error(
                    deployment_tuple,
                    this_deployment_subscription,
                    set(loaded_data["service_principals"].keys()),
                    "service principals",
                )

            if (
                ("static_secrets" in loaded_data)
                and (loaded_data["static_secrets"])
                and (
                    this_deployment_subscription
                    not in loaded_data["static_secrets"]
                )
            ):
                # a deployment subscription must be defined in static_secrets
                PipelineConfiguration.__report_deployment_subscription_error(
                    deployment_tuple,
                    this_deployment_subscription,
                    set(loaded_data["static_secrets"].keys()),
                    "static secrets",
                )

    @pydantic.root_validator()
    def check_deployments(cls: pydantic.BaseModel, loaded_data: dict) -> dict:
        """Cross-check loaded deployment data."""
        deployment_names = set(
            loaded_data["deployments"].deployment_tuples.keys()
        )
        for this_name in deployment_names:
            typing.cast("PipelineConfiguration", cls).__check_deployment_tuple(
                this_name, loaded_data
            )
            typing.cast(
                "PipelineConfiguration", cls
            ).__check_deployment_subscriptions(this_name, loaded_data)

        return loaded_data

    @pydantic.root_validator()
    def check_service_principals(
        cls: pydantic.BaseModel, loaded_data: dict
    ) -> dict:
        """Validate service principal data."""
        if loaded_data["service_principals"]:
            PipelineConfiguration.__check_subscriptions_in_subscriptions(
                loaded_data, "service_principals"
            )
        return loaded_data

    @staticmethod
    def __check_subscriptions_in_subscriptions(
        loaded_data: dict, data_name: str
    ) -> None:
        """Check that subscriptions names are defined in subscriptions."""
        bad_subscriptions = [
            "{0}".format(name)
            for name in loaded_data[data_name].keys()
            if name not in loaded_data["subscriptions"].keys()
        ]
        if any(bad_subscriptions):
            message = f"{data_name} subscriptions not defined in subscriptions"
            log.error("{0}, {1}".format(message, str(bad_subscriptions)))
            raise PipelineConfigurationError(message)

    @pydantic.root_validator()
    def check_static_secrets(
        cls: pydantic.BaseModel, loaded_data: dict
    ) -> dict:
        """Validate static secret data."""
        if loaded_data["static_secrets"]:
            PipelineConfiguration.__check_subscriptions_in_subscriptions(
                loaded_data, "static_secrets"
            )
        return loaded_data

    @pydantic.root_validator()
    def check_subscriptions(cls: pydantic.BaseModel, loaded_data: dict) -> dict:
        """Validate subscriptions data."""
        bad_tenants = [
            "{0}.{1}".format(name, data.tenant)
            for name, data in loaded_data["subscriptions"].items()
            if data.tenant not in loaded_data["tenants"]
        ]
        if any(bad_tenants):
            message = "Bad tenant(s) in subscription"
            log.error(
                "{0}, {1}, {2}".format(
                    message, str(bad_tenants), str(loaded_data["tenants"])
                )
            )
            raise PipelineConfigurationError(message)
        return loaded_data



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#
# https://gitlab.com/ci-cd-devops/build_harness/-/blob/main/build_harness/_utility.py

"""General support for build harness implementation."""

import asyncio
import dataclasses
import locale
import logging
import pathlib
import shutil
import subprocess
import sys
import typing

from ._exceptions import CommandError

log = logging.getLogger(__name__)

CommandArgs = typing.List[str]


@dataclasses.dataclass()
class CapturedStreams:
    """Encapsulate output and error streams captured from a process."""

    out: str
    error: str


def detect_venv_command(command_name: str) -> pathlib.Path:
    """Detect a command in the same venv as the current utility."""
    venv_path = pathlib.Path(sys.argv[0]).parent.absolute()

    expected_command_path = venv_path / command_name
    if expected_command_path.is_file():
        result = expected_command_path
    else:
        # assume command in the PATH when run...
        found = shutil.which(command_name)
        if found:
            result = pathlib.Path(found)
        else:
            raise CommandError(
                "Command not in users path or venv, {0}".format(command_name)
            )

    return result


def run_command(
    command: CommandArgs, enable_logging: bool = False, **kwargs: typing.Any
) -> subprocess.CompletedProcess:
    """
    Run a system command using ``subprocess.run``.

    Args:
        command: List of command and arguments.
        enable_logging: Enable command and argument logging (default disabled)
        **kwargs: Optional arguments passed through to ``subprocess.run``.

    Returns:
        Subprocess results.
    """
    if enable_logging:
        # WARNING: logging is _disabled_ by default to prevent "default"
        # leakage of secrets into logs via command arguments
        log.debug("command to run, {0}".format(str(command)))
        log.debug("command arguments, {0}".format(str(kwargs)))
    result = subprocess.run(command, **kwargs)

    return result


async def run_async_command(
    command: CommandArgs, enable_logging: bool = False
) -> CapturedStreams:
    """
    Run an external command asynchronously.

    Command logging is disabled by default to help prevent accidental
    credential leakage to logs when commands may necessarily contain
    sensitive data.

    Args:
        command: Command to be executed.
        enable_logging: Enable command and argument logging (default disabled)

    Returns:
        Any output or error streams captured from the process.
    Raises:
        CommandError: if the executed command returns non-zero exit status.
    """
    if enable_logging:
        # WARNING: logging is _disabled_ by default to prevent "default"
        # leakage of secrets into logs via command arguments
        log.debug("command to run, {0}".format(str(command)))

    this_encoding = sys.getfilesystemencoding()
    log.debug("sys.getdefaultencoding(), {0}".format(sys.getdefaultencoding()))
    log.debug(
        "locale.getpreferredencoding(), {0}".format(
            locale.getpreferredencoding()
        )
    )
    log.debug(
        "sys.getfilesystemencoding(), {0}".format(sys.getfilesystemencoding())
    )
    this_process = await asyncio.create_subprocess_exec(
        *command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await this_process.communicate()
    log.debug(
        "run_async_command stdout, {0}".format(stdout.decode(this_encoding))
    )
    log.debug(
        "run_async_command stderr, {0}".format(stderr.decode(this_encoding))
    )
    result = CapturedStreams(
        out=stdout.decode(this_encoding), error=stderr.decode(this_encoding)
    )
    if this_process.returncode != 0:
        log.debug(f"command std out, {result.out}")
        raise CommandError(
            "External command run did not exit cleanly, {0}".format(
                result.error
            )
        )
    return result



#!python3
# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Release id utility."""

import enum
import re

from ..patterns import SEMANTIC_VERSION_GITREF, TAG_PREFIX
from ..release import ReleaseState
from ._cd_release_state import identify_release_state
from ._exceptions import ReleaseIdentityError, ReleaseStateError


@enum.unique
class ExitState(enum.Enum):
    """Release id utility CLI exit states."""

    UNKNOWN = 100
    MISSING_GITREF = 101
    GITREF_PARSE_FAILURE = 102


REGEX_TABLE = {
    ReleaseState.qa: r"^{0}(?P<id>{1}-alpha\.\d+)$".format(
        TAG_PREFIX, SEMANTIC_VERSION_GITREF
    ),
    ReleaseState.stg: r"^{0}(?P<id>{1}-beta\.\d+)$".format(
        TAG_PREFIX, SEMANTIC_VERSION_GITREF
    ),
    ReleaseState.prd: r"^{0}(?P<id>{1})$".format(
        TAG_PREFIX, SEMANTIC_VERSION_GITREF
    ),
}

ERROR_TEXT = "Unable to match git reference to any release id, {0}"


def identify_release_id(git_ref: str) -> str:
    """
    Parse git reference to extract release identity.

    Args:
        git_ref: git reference to parse

    Returns:
        Extracted release identity.
    Raises:
        ReleaseIdentityError: If no release id can be inferred.
    """
    try:
        state = identify_release_state(git_ref)

        found_id = False
        id_found = None
        if state in REGEX_TABLE:
            match_result = re.match(REGEX_TABLE[state], git_ref)
            if match_result is not None:
                found_id = True
                id_found = match_result.group("id")
        elif state in [ReleaseState.ftr, ReleaseState.dev]:
            # PEP-440 style "-post" convention should be valid for
            # https://semver.org/ compliant systems.
            found_id = True
            # WARNING: Placeholder! Needs to apply `git describe` to repo and
            # extract the correct attribute.
            id_found = "0.0.0-post.000"

        if (not found_id) or (not id_found):
            raise ReleaseIdentityError(ERROR_TEXT.format(git_ref))
    except ReleaseStateError as e:
        raise ReleaseIdentityError(ERROR_TEXT.format(git_ref)) from e

    return id_found



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""File maintainer utility."""

import asyncio
import datetime
import enum
import functools
import logging
import os
import pathlib
import shutil
import sys
import typing

import click

from foodx_devops_tools._version import acquire_version

from ._declarations import (
    DEFAULT_CONSOLE_LOGGING_ENABLED,
    DEFAULT_FILE_LOGGING_DISABLED,
    DEFAULT_LOG_LEVEL,
    VALID_LOG_LEVELS,
)
from ._logging import LoggingState

log = logging.getLogger(__name__)

DEFAULT_LOG_FILE = pathlib.Path("file_maintainer.log")
MAX_DELETE_ITEMS = 10


@enum.unique
class ExitState(enum.Enum):
    """Maintain console utility exit states."""

    UNKNOWN = enum.auto()


def _get_filesystem_capacity_used(directory: pathlib.Path) -> int:
    """
    Calculate the used capacity of the filesystem in percent.

    Returns:
        Percentage of capacity used.
    """
    stats = os.statvfs(directory)

    size_bytes = stats.f_frsize * stats.f_blocks
    available_bytes = stats.f_frsize * stats.f_bavail

    capacity_used_percent = int((1 - (available_bytes / size_bytes)) * 100)

    log.info("filesystem size (bytes), {0}".format(size_bytes))
    log.info("filesystem available size (bytes), {0}".format(size_bytes))

    log.info("filesystem capacity used (%), {0}".format(capacity_used_percent))

    return capacity_used_percent


T = typing.TypeVar("T", bound="RunMonitor")


class RunMonitor:
    """Maintain the iterations for when persistence is not required."""

    def __init__(
        self: T, persist_interval_minutes: typing.Optional[int]
    ) -> None:
        """Construct ``RunMonitor`` object."""
        self.__persist_interval_minutes = persist_interval_minutes
        self.__loop_count = 0

    def keep_running(self: T) -> bool:
        """Determine when to keep a loop running."""
        # if persistence is specified the loop always keeps running
        result = True
        if (self.__persist_interval_minutes is None) and (
            self.__loop_count == 0
        ):
            # increment the counter so that the next time around the loop the
            # iteration will be cancelled
            self.__loop_count += 1
            log.info("no persistence. exiting after this iteration")
        elif self.__persist_interval_minutes is None:
            # cancel iterating on the loop because the loop counter is non-zero
            log.info("cancelling iteration due to no persistence")
            result = False
        else:
            log.debug("persistence specified")

        return result

    async def sleep(self: T) -> None:
        """Sleep for an interval if necessary."""
        # only sleep when the persistence interval has been specified
        if self.__persist_interval_minutes is not None:
            duration_seconds = self.__persist_interval_minutes * 60
            log.debug(
                "sleeping for interval, {0} (seconds)".format(duration_seconds)
            )

            await asyncio.sleep(duration_seconds)
        else:
            log.info("not sleeping because no persistence specified")


async def _do_delete(this_dir: pathlib.Path) -> None:
    log.info("deleting subdirectory, {0}".format(this_dir))

    this_partial = functools.partial(pathlib.Path, ignore_errors=True)

    loop = asyncio.get_event_loop()
    await loop.run_in_executor(None, shutil.rmtree, this_partial(this_dir))


async def _clean_filesystem(directory: pathlib.Path) -> None:
    subdirectories = [x for x in directory.iterdir() if x.is_dir()]
    subdirectory_ages = {
        x: datetime.datetime.fromtimestamp(x.stat().st_mtime)
        for x in subdirectories
    }
    # NOTE: this only works Python >=3.7 when dict insertion order
    # preservation was implemented.
    subdirectories_by_age = {
        k: v
        for k, v in sorted(
            subdirectory_ages.items(), key=lambda item: item[1], reverse=True
        )
    }

    number_items = MAX_DELETE_ITEMS
    if len(subdirectories_by_age) <= MAX_DELETE_ITEMS:
        number_items = len(subdirectories_by_age) - 1

    delete_items = list(subdirectories_by_age.keys())[0:number_items]

    click.echo("deleting subdirectories to improve available space")
    await asyncio.gather(*[_do_delete(x) for x in delete_items])


async def _run_maintainer(
    directory: pathlib.Path,
    persist_interval_minutes: typing.Optional[int],
    threshold_percent: int,
) -> None:
    this_iteration = RunMonitor(persist_interval_minutes)
    while this_iteration.keep_running():
        capacity_used_percent = _get_filesystem_capacity_used(directory)

        if capacity_used_percent >= threshold_percent:
            click.echo(
                "capacity above threshold, {0} ({1})".format(
                    capacity_used_percent, threshold_percent
                )
            )
            await _clean_filesystem(directory)

        await this_iteration.sleep()


@click.command()
@click.version_option(version=acquire_version())
@click.argument(
    "directory",
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
)
@click.option(
    "--log-enable-console",
    "enable_console_log",
    default=DEFAULT_CONSOLE_LOGGING_ENABLED,
    help="Log to console.",
    is_flag=True,
)
@click.option(
    "--log-disable-file",
    "disable_file_log",
    default=DEFAULT_FILE_LOGGING_DISABLED,
    help="Disable file logging.",
    is_flag=True,
)
@click.option(
    "--log-level",
    "log_level",
    default=DEFAULT_LOG_LEVEL,
    help="Select logging level to apply to all enabled log sinks.",
    show_default=True,
    type=click.Choice(VALID_LOG_LEVELS, case_sensitive=False),
)
@click.option(
    "--persist-interval",
    "persist_interval_minutes",
    default=None,
    help="Persist indefinitely checking the directory on the specified "
    "interval (minutes).",
    show_default=True,
    type=int,
)
@click.option(
    "--threshold",
    default=80,
    help="Filesystem threshold to trigger clean up.",
    show_default=True,
    type=int,
)
def file_maintainer(
    directory: pathlib.Path,
    disable_file_log: bool,
    enable_console_log: bool,
    log_level: str,
    persist_interval_minutes: typing.Optional[int],
    threshold: int,
) -> None:
    """
    Delete old sub-directories from the specified directory.

    By default the utility will check the specified directory, erase relevant
    subdirectories and exit. Use the ``--persist-interval`` option for the
    utility to run indefinitely, checking the directory periodically on the
    specified interval.

    DIRECTORY:  The directory to monitor.
    """
    try:
        # currently no need to change logging configuration at run time,
        # so no need to preserve the object.
        LoggingState(
            disable_file_logging=disable_file_log,
            enable_console_logging=enable_console_log,
            log_level_text=log_level,
            default_log_file=DEFAULT_LOG_FILE,
        )

        asyncio.run(
            _run_maintainer(directory, persist_interval_minutes, threshold)
        )

        click.echo("maintainer exiting due to task completion")
    except asyncio.exceptions.CancelledError:
        log.warning("exiting due to cancellation")
        # async CancelledError exceptions should usually be re-raised
        # https://docs.python.org/3.8/library/asyncio-exceptions.html#asyncio.CancelledError
        raise
    except Exception as e:
        click.echo(
            "failed with unexpected error (see log for "
            "details), {0}".format(str(e)),
            err=True,
        )
        log.exception("unexpected error, {0}".format(str(e)))
        sys.exit(ExitState.UNKNOWN.value)


def flit_entry() -> None:
    """Flit script entry function for ``file-maintainer`` utility."""
    file_maintainer()



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio
import contextlib
import copy
import json
import pathlib
import typing
import uuid

import pytest

from foodx_devops_tools.puff.arm import _save_parameter_file


class TestSaveParameterFile:
    @pytest.mark.asyncio
    async def test_create_clean(self, tmp_path_factory):
        data = {
            "p1": "v1",
            "p2": "v2",
        }
        content = await self._do_create_test(data, tmp_path_factory)
        assert "p1" in content["parameters"]
        assert "p2" in content["parameters"]
        assert content["parameters"]["p1"]["value"] == "v1"
        assert content["parameters"]["p2"]["value"] == "v2"

    def _check_basic_content(self, this_file: pathlib.Path) -> dict:
        assert this_file.exists()
        with this_file.open("r") as f:
            content = json.load(f)

        assert all(
            [x in content for x in ["$schema", "contentVersion", "parameters"]]
        )
        return content

    async def _do_create_test(
        self,
        parameter_data: typing.Optional[dict],
        path_factory,
        is_pretty: bool = False,
    ) -> dict:
        this_directory = path_factory.mktemp(str(uuid.uuid4()))
        this_file = this_directory / "some_file"
        assert not this_file.exists()

        await _save_parameter_file(this_file, parameter_data, is_pretty)

        return self._check_basic_content(this_file)

    @pytest.mark.asyncio
    async def test_reference_clean(self, tmp_path_factory):
        data = {
            "p1": "v1",
            "p2": "v2",
            "p3": {
                "reference": {
                    "keyVault": {"id": "idv"},
                    "secretName": "sn",
                }
            },
        }
        content = await self._do_create_test(data, tmp_path_factory)
        assert "p1" in content["parameters"]
        assert "p2" in content["parameters"]
        assert content["parameters"]["p1"]["value"] == "v1"
        assert content["parameters"]["p2"]["value"] == "v2"
        assert (
            content["parameters"]["p3"]["reference"]["keyVault"]["id"] == "idv"
        )
        assert content["parameters"]["p3"]["reference"]["secretName"] == "sn"

    @pytest.mark.asyncio
    async def test_nonstring_clean(self, tmp_path_factory):
        """all value parameters must be string"""
        data = {
            "p1": "v1",
            "p2": "v2",
            "p3": 3.14,
            "p4": 10,
            "p5": False,
            "p6": [1, 2, 3],
            "p7": {
                "reference": {"keyVault": {"id": "ivd"}, "secretName": "sn"}
            },
            "p8": {"value": {"k1": "kv1", "k2": "kv2"}},
        }
        content = await self._do_create_test(data, tmp_path_factory)
        assert "p1" in content["parameters"]
        assert "p2" in content["parameters"]
        assert content["parameters"]["p1"]["value"] == "v1"
        assert content["parameters"]["p2"]["value"] == "v2"
        assert content["parameters"]["p3"]["value"] == 3.14
        assert content["parameters"]["p4"]["value"] == 10
        assert content["parameters"]["p5"]["value"] is False
        assert content["parameters"]["p6"]["value"] == [1, 2, 3]
        assert content["parameters"]["p7"]["reference"] == {
            "keyVault": {"id": "ivd"},
            "secretName": "sn",
        }
        assert content["parameters"]["p8"]["value"] == {
            "k1": "kv1",
            "k2": "kv2",
        }

    @pytest.mark.asyncio
    async def test_none_parameters(self, tmp_path_factory):
        await self._do_create_test(None, tmp_path_factory)

    @pytest.mark.asyncio
    async def test_empty_parameters(self, tmp_path_factory):
        await self._do_create_test(dict(), tmp_path_factory)

    @pytest.mark.asyncio
    async def test_pretty(self, mock_context, mocker):
        mock_context("foodx_devops_tools.puff.arm.aiofiles.open")
        mock_dump = mocker.patch("foodx_devops_tools.puff.arm.json.dumps")

        this_directory = pathlib.Path("some/path")
        this_file = this_directory / "some_file"

        await _save_parameter_file(this_file, dict(), True)

        mock_dump.assert_called_once_with(mocker.ANY, sort_keys=True, indent=2)

    @pytest.mark.asyncio
    async def test_not_pretty(self, mock_context, mocker):
        mock_context("foodx_devops_tools.puff.arm.aiofiles.open")
        mock_dump = mocker.patch("foodx_devops_tools.puff.arm.json.dumps")

        this_directory = pathlib.Path("some/path")
        this_file = this_directory / "some_file"

        await _save_parameter_file(this_file, dict(), False)

        mock_dump.assert_called_once_with(mocker.ANY)


class MockFileStream:
    def __init__(self, target: pathlib.Path, mode: str = "r"):
        self.sleep_intervals = {0.1, 1}

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    async def write(self, stuff: str):
        pass


class TestConcurrentTargets:
    @pytest.mark.asyncio
    async def test_concurrent_save(self, mocker):
        mock_dump = mocker.patch("foodx_devops_tools.puff.arm.json.dumps")
        mocker.patch(
            "foodx_devops_tools.puff.arm.aiofiles.open",
            side_effect=MockFileStream,
        )
        data = [
            {
                "expected": {
                    "$schema": "https://schema.management.azure.com/schemas/2019-04-01"
                    "/deploymentParameters.json#",
                    "contentVersion": "1.0.0.0",
                    "parameters": {"av": {"value": "3.14"}},
                },
                "path": pathlib.Path("a"),
                "parameters": {"av": 3.14},
            },
            {
                "expected": {
                    "$schema": "https://schema.management.azure.com/schemas/2019-04-01"
                    "/deploymentParameters.json#",
                    "contentVersion": "1.0.0.0",
                    "parameters": {"bv": {"value": "something"}},
                },
                "path": pathlib.Path("b"),
                "parameters": {"bv": "something"},
            },
        ]
        await asyncio.gather(
            *[
                _save_parameter_file(x["path"], x["parameters"], True)
                for x in data
            ]
        )

        a = mock_dump.mock_calls[0].args[0]
        b = mock_dump.mock_calls[1].args[0]

        # Ensure that the parameters sub-dict is unique in each concurrent call.
        assert id(a["parameters"]) != id(b["parameters"])
        assert a["parameters"] != b["parameters"]



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Azure ARM template parameter generation."""

import copy
import json
import logging
import os
import pathlib
import typing

import aiofiles
import aiofiles.os
import click
import pydantic
import ruamel.yaml
from deepmerge import always_merger  # type: ignore

from ._exceptions import ArmTemplateError
from ._header import ARMTEMPLATE_PARAMETERS_HEADER
from ._puff_parameters import PuffParameterModel

log = logging.getLogger(__name__)

YamlData = typing.Dict[str, typing.Any]

BASE_PARAMETER_EXCLUDE_LABELS = ["default", "environments", "name", "services"]


async def load_yaml(path: pathlib.Path) -> YamlData:
    """
    Asynchronously load yaml file data.

    Args:
        path: Path to YAML file to load.

    Returns:
        Loaded YAML data.
    Raises:
        FileNotFoundError: If file cannot be loaded.
        IsADirectoryError: If path is a directory instead of a file.
    """
    try:
        yaml = ruamel.yaml.YAML(typ="safe")
        async with aiofiles.open(str(path), mode="r") as f:
            content = await f.read()
        yaml_data = yaml.load(content)
        log.debug("{0}, {1}".format(path, str(yaml_data)))

        if not yaml_data:
            log.warning("Empty YAML file, {0}".format(str(path)))

        return yaml_data
    except FileNotFoundError:
        log.error("File not found, {0}".format(str(path)))
        raise
    except IsADirectoryError:
        log.error("Path is a directory, {0}".format(str(path)))
        raise
    except Exception:
        log.error("Unexpected error with path, {0}".format(str(path)))
        raise


def _remove_keys(
    parameters: typing.Dict[str, typing.Any], exclude_labels: typing.List[str]
) -> dict:
    """
    Remove keys from a dictionary without changing the original.

    Attempts to remove keys that don't exist in the dictinary are silently
    ignored.

    Args:
        parameters: Dictionary to be adjusted.

    Returns:
        Modified dictionary.
    """
    this_copy = parameters.copy()
    for k in exclude_labels:
        try:
            del this_copy[k]
        except KeyError:
            # silently ignore missing keys by suppressing this exception
            pass

    return this_copy


async def _save_parameter_file(
    target_path: pathlib.Path,
    parameter_data: typing.Optional[dict],
    is_pretty: bool,
) -> None:
    """
    Create ARM template parameter file.

    Args:
        target_path: File path to be created or deleted.
        parameter_data: Parameters to be applied to data.
        is_pretty: Create nicely formatted JSON for humans.
    """
    # WARNING: The header template dict contains a dict in a field and
    # dict.copy() is shallow so care needs to be taken to ensure all
    # concurrent copies are unique due to Python reference-copy default.
    generated_parameters = copy.deepcopy(ARMTEMPLATE_PARAMETERS_HEADER)
    if parameter_data:
        parameters = dict()
        for item_key, value in parameter_data.items():
            if isinstance(value, dict):
                # assume the dict is valid ARM template parameter structure.
                # the dict SHOULD be an ARM template parameter reference:
                # https://docs.microsoft.com/en-us/azure/azure-resource-manager/templates/parameter-files#parameter-file
                parameters[item_key] = value
            elif isinstance(value, (bool, float, int, str, list)):
                parameters[item_key] = {
                    "value": value,
                }
            else:
                # assume the value is convertible to str
                parameters[item_key] = {
                    "value": str(value),
                }
        generated_parameters["parameters"] = parameters
    dump_arguments: dict = dict()
    if is_pretty:
        dump_arguments = {"sort_keys": True, "indent": 2}

    if not target_path.parent.exists():
        os.makedirs(str(target_path.parent), exist_ok=True)
    async with aiofiles.open(target_path, mode="w") as f:
        log.info("saving parameter file, {0}".format(target_path))
        await f.write(json.dumps(generated_parameters, **dump_arguments))


async def _delete_parameter_file(
    target_path: pathlib.Path,
) -> None:
    """
    Delete generated ARM template parameter file.

    Args:
        target_path: File path to be created or deleted.
    """
    if target_path.exists():
        log.warning("deleting parameter file, {0}".format(target_path))
        await aiofiles.os.remove(target_path)


async def _do_file_actions(
    is_delete_action: bool,
    target_path: pathlib.Path,
    parameter_data: dict,
    is_pretty: bool,
) -> None:
    """
    Create or delete generated ARM template parameter files.

    Args:
        is_delete_action: Signal deleting or creating parameter files.
        target_path: Directory to store or delete parameter files.
        parameter_data: Data for each ARM template parameter file.
        is_pretty: Create nicely formatted JSON for humans.
    """
    for key, values in parameter_data.items():
        this_path = target_path / ".".join([key, "json"])
        if is_delete_action:
            await _delete_parameter_file(this_path)
        else:
            await _save_parameter_file(this_path, values, is_pretty)


def _linearize_name(base_data: dict, filename: str) -> dict:
    cleaned = _remove_keys(base_data, ["name"])
    this_name = base_data.get("name", filename)
    if not this_name:
        raise RuntimeError("Invalid empty name")

    linearized = {
        this_name: cleaned,
    }
    return linearized


def _merge_default_name(base_data: dict) -> dict:
    this_data = copy.deepcopy(base_data)
    for key, value in this_data.items():
        if "default" in value:
            new_parameters = value["default"]
            cleaned = _remove_keys(value, ["default"])

            updated = always_merger.merge(cleaned, new_parameters)

            this_data[key] = updated

    return this_data


def _iterate_iterable(
    cleaned: dict, base_key: str, this_data: dict, singular: str
) -> dict:
    linearized = dict()
    for iterable_key, iterable_data in this_data.items():
        if not iterable_key:
            raise RuntimeError("Invalid empty {0} name".format(singular))

        if iterable_data:
            this_value = always_merger.merge(
                copy.deepcopy(cleaned), iterable_data
            )
        else:
            this_value = copy.deepcopy(cleaned)

        this_value[singular] = iterable_key
        linearized["{0}.{1}".format(base_key, iterable_key)] = this_value

    return linearized


def _linearize_iterable(base_data: dict, plural: str, singular: str) -> dict:
    linearized = dict()
    for base_key, base_value in base_data.items():
        this_data = (
            copy.deepcopy(base_value[plural])
            if plural in base_value
            else dict()
        )
        cleaned = _remove_keys(copy.deepcopy(base_value), [plural])
        if this_data:
            iterated = _iterate_iterable(cleaned, base_key, this_data, singular)
            linearized.update(iterated)
        else:
            linearized[base_key] = cleaned

    return linearized if linearized else base_data.copy()


def _linearize_services(base_data: dict) -> dict:
    plural = "services"
    singular = "service"
    return _linearize_iterable(base_data, plural, singular)


def _linearize_environments(base_data: dict) -> dict:
    plural = "environments"
    singular = "environment"
    return _linearize_iterable(base_data, plural, singular)


def _merge_list_of_dict(this_data: typing.List[dict]) -> dict:
    result: dict = dict()
    for item in this_data:
        if isinstance(item, dict):
            for k, v in item.items():
                if k in result:
                    result[k].update(v if v is not None else dict())
                else:
                    result[k] = v if v is not None else dict()
        else:
            log.error("ignoring non-dict list entry, {0}".format(str(item)))
    return result


def _linearize_regions(base_data: dict) -> dict:
    plural = "regions"
    singular = "region"
    linearized = dict()
    for base_key, base_value in base_data.items():
        this_data = (
            copy.deepcopy(base_value[plural])
            if plural in base_value
            else list()
        )
        cleaned = _remove_keys(base_value, [plural])
        # special handling of the region list unique to regions
        if this_data:
            merged_list = _merge_list_of_dict(this_data)
            iterated = _iterate_iterable(
                cleaned, base_key, merged_list, singular
            )
            linearized.update(iterated)
        else:
            linearized[base_key] = cleaned

    return linearized if linearized else copy.deepcopy(base_data)


def _linearize_parameters(yaml_data: dict, file_name: str) -> dict:
    level0_merge = _linearize_name(yaml_data, file_name)

    # a hack to enable a variable called "name" to be specified.
    default_merged = _merge_default_name(level0_merge)

    level1_merge = _linearize_services(default_merged)
    level2_merge = _linearize_environments(level1_merge)
    level3_merge = _linearize_regions(level2_merge)

    return level3_merge


async def do_arm_template_parameter_action(
    puff_file_path: pathlib.Path,
    output_dir: typing.Optional[pathlib.Path],
    is_delete_action: bool,
    is_pretty: bool,
) -> None:
    """
    Generate ARM template parameter files from puff YAML file.

    The generated files are always output to the same directory as the source
    puff file.

    Args:
        puff_file_path:     Path to source YAML parameter file.
        output_dir:         Optional path to directory for storing output files.
        is_delete_action:   True if files should be deleted instead of created.
        is_pretty:          Create nicely formatted JSON for humans.
    """
    if output_dir:
        target_path = output_dir
    else:
        target_path = puff_file_path.parent

    click.echo("loading, {0}".format(puff_file_path))
    yaml_data = await load_yaml(puff_file_path)
    try:
        # for now, just use the pydantic model to validate the YAML data.
        PuffParameterModel.parse_obj(yaml_data)

        file_name = pathlib.Path(puff_file_path.stem).name
        merged_data = _linearize_parameters(yaml_data, file_name)
        await _do_file_actions(
            is_delete_action, target_path, merged_data, is_pretty
        )
    except pydantic.ValidationError as e:
        raise ArmTemplateError(
            "Puff parameter validation failed, "
            "{0}. {1}".format(puff_file_path, str(e))
        ) from e



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.


from foodx_devops_tools.puff.arm import _linearize_services


class TestLinearizedServices:
    def test_services_environments(self):
        mock_base = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "environments": {
                    "e1": {
                        "k1": "be1k1",
                        "k2": "be1k2",
                    },
                },
                "services": {
                    "s1": {
                        "p2": "s1p2",
                        "environments": {
                            "e1": {
                                "k1": "s1e1k1",
                            },
                        },
                    },
                },
            },
        }
        expected_result = {
            "some-file.s1": {
                "service": "s1",
                "p1": "bp1",
                "p2": "s1p2",
                "environments": {
                    "e1": {
                        "k1": "s1e1k1",
                        "k2": "be1k2",
                    },
                },
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_multiple_services(self):
        mock_base = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "services": {
                    "s1": {
                        "p2": "s1p2",
                    },
                    "s2": {
                        "p2": "s2p2",
                    },
                },
            },
        }
        expected_result = {
            "some-file.s1": {
                "service": "s1",
                "p1": "bp1",
                "p2": "s1p2",
            },
            "some-file.s2": {
                "service": "s2",
                "p1": "bp1",
                "p2": "s2p2",
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_services_environments2(self):
        mock_base = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "environments": {
                    "e1": {
                        "k1": "be1k1",
                        "k2": "be1k2",
                    },
                },
                "services": {
                    "s1": {
                        "p2": "s1p2",
                        "environments": {
                            "e1": {
                                "k1": "s1e1k1",
                            },
                        },
                    },
                },
            },
        }
        expected_result = {
            "some-file.s1": {
                "service": "s1",
                "p1": "bp1",
                "p2": "s1p2",
                "environments": {
                    "e1": {
                        "k1": "s1e1k1",
                        "k2": "be1k2",
                    },
                },
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_empty_services(self):
        mock_base = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "environments": {
                    "e1": {
                        "k1": "be1k1",
                        "k2": "be1k2",
                    },
                },
                "services": dict(),
            },
        }
        expected_result = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "environments": {
                    "e1": {
                        "k1": "be1k1",
                        "k2": "be1k2",
                    },
                },
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_empty_service(self):
        mock_base = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "services": {
                    "s1": dict(),
                },
            },
        }
        expected_result = {
            "some-file.s1": {
                "service": "s1",
                "p1": "bp1",
                "p2": "bp2",
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_none_service(self):
        mock_base = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "services": {
                    "s1": None,
                },
            },
        }
        expected_result = {
            "some-file.s1": {
                "service": "s1",
                "p1": "bp1",
                "p2": "bp2",
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_no_services_no_environments(self):
        mock_base = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
                "services": dict(),
            },
        }
        expected_result = {
            "some-file": {
                "p1": "bp1",
                "p2": "bp2",
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_no_services_no_environments_empty_base(self):
        mock_base = {
            "some-file": dict(),
        }
        expected_result = mock_base
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_services_empty_base(self):
        mock_base = {
            "some-file": {
                "services": {
                    "s1": {
                        "p2": "s1p2",
                        "environments": {
                            "e1": {
                                "k1": "s1e1k1",
                            },
                        },
                    },
                },
            },
        }
        expected_result = {
            "some-file.s1": {
                "service": "s1",
                "p2": "s1p2",
                "environments": {
                    "e1": {
                        "k1": "s1e1k1",
                    },
                },
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result

    def test_complex1_clean(self):
        mock_base = {
            "filestub": {
                "p1": "bp1",
                "bp2": "bp2",
                "environments": {
                    "e1": {
                        "p1": "e1p1",
                        "e1p2": "e1p2",
                        "regions": [
                            {
                                "r1": {
                                    "e1r1p2": "e1r1p2",
                                },
                            }
                        ],
                    },
                },
                "services": {
                    "s1": {
                        "s1p2": "s1p2",
                        "environments": {
                            "e1": {
                                "s1e1p2": "s1e1p2",
                                "regions": [
                                    {
                                        "r1": {
                                            "p1": "s1e1r1p1",
                                        },
                                    },
                                ],
                            },
                        },
                    },
                    "s2": {
                        "s2p2": "s2p2",
                        "environments": {
                            "e1": {
                                "s2e1p2": "s2e1p2",
                                "regions": [
                                    {
                                        "r1": {
                                            "p1": "s2e1r1p1",
                                        },
                                    },
                                ],
                            },
                        },
                    },
                },
            },
        }
        expected_result = {
            "filestub.s1": {
                "service": "s1",
                "environments": {
                    "e1": {
                        "p1": "e1p1",
                        "e1p2": "e1p2",
                        "s1e1p2": "s1e1p2",
                        "regions": [
                            {"r1": {"e1r1p2": "e1r1p2"}},
                            {"r1": {"p1": "s1e1r1p1"}},
                        ],
                    },
                },
                "p1": "bp1",
                "bp2": "bp2",
                "s1p2": "s1p2",
            },
            "filestub.s2": {
                "service": "s2",
                "environments": {
                    "e1": {
                        "p1": "e1p1",
                        "e1p2": "e1p2",
                        "s2e1p2": "s2e1p2",
                        "regions": [
                            {"r1": {"e1r1p2": "e1r1p2"}},
                            {
                                "r1": {
                                    "p1": "s2e1r1p1",
                                },
                            },
                        ],
                    },
                },
                "p1": "bp1",
                "bp2": "bp2",
                "s2p2": "s2p2",
            },
        }
        result = _linearize_services(mock_base)

        assert result == expected_result



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy
import logging

import pytest

from foodx_devops_tools.pipeline_config import (
    ClientsDefinition,
    DeploymentsDefinition,
    FramesDefinition,
    PipelineConfiguration,
    ReleaseStatesDefinition,
    ServicePrincipals,
    StaticSecrets,
    SubscriptionsDefinition,
    SystemsDefinition,
    TenantsDefinition,
)
from tests.ci.support.pipeline_config import MOCK_RESULTS

log = logging.getLogger(__name__)


@pytest.fixture()
def mock_loads(mocker):
    def _apply(mock_data: PipelineConfiguration) -> None:
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_clients",
            return_value=ClientsDefinition.parse_obj(
                {"clients": mock_data.clients}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_deployments",
            return_value=DeploymentsDefinition.parse_obj(
                {"deployments": mock_data.deployments}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_frames",
            return_value=FramesDefinition.parse_obj(
                {"frames": mock_data.frames}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_release_states",
            return_value=ReleaseStatesDefinition.parse_obj(
                {"release_states": mock_data.release_states}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_service_principals",
            return_value=ServicePrincipals.parse_obj(
                {"service_principals": mock_data.service_principals}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_static_secrets",
            return_value=StaticSecrets.parse_obj(
                {"static_secrets": mock_data.static_secrets}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_subscriptions",
            return_value=SubscriptionsDefinition.parse_obj(
                {"subscriptions": mock_data.subscriptions}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_systems",
            return_value=SystemsDefinition.parse_obj(
                {"systems": mock_data.systems}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_tenants",
            return_value=TenantsDefinition.parse_obj(
                {"tenants": mock_data.tenants}
            ),
        )

    return _apply


@pytest.fixture()
def mock_results():
    results_copy = copy.deepcopy(MOCK_RESULTS)
    pipeline_config = PipelineConfiguration.parse_obj(results_copy)

    return pipeline_config



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Deployment status data model."""

import asyncio
import copy
import dataclasses
import enum
import logging
import typing

import click

log = logging.getLogger(__name__)

DEFAULT_MONITOR_SLEEP_SECONDS = 10


@dataclasses.dataclass
class DeploymentState:
    """Signal state of a deployment."""

    @enum.unique
    class ResultType(enum.Enum):
        """Deployment state enumeration."""

        cancelled = enum.auto()
        failed = enum.auto()
        in_progress = enum.auto()
        pending = enum.auto()
        skipped = enum.auto()
        success = enum.auto()

    COMPLETED_OK = {
        ResultType.skipped,
        ResultType.success,
    }
    COMPLETED_RESULTS = {
        ResultType.cancelled,
        ResultType.failed,
        ResultType.skipped,
        ResultType.success,
    }

    code: ResultType
    message: typing.Optional[str] = None


def all_success(values: typing.List[DeploymentState]) -> bool:
    """
    Evaluate if a list of deployment states are all succeeded.

    NOTE: "skipped" is also considered a success here.
    """
    result = all([x.code in DeploymentState.COMPLETED_OK for x in values])
    return result


def all_completed(values: typing.List[DeploymentState]) -> bool:
    """
    Evaluate if a list of deployment states are all completed.

    "All completed" is defined here as any one of the deployment states that
    can be interpreted as the process having stopped for some reason;
    success, failed or cancelled.
    """
    result = all([x.code in DeploymentState.COMPLETED_RESULTS for x in values])
    return result


def any_completed_dirty(values: typing.List[DeploymentState]) -> bool:
    """Evaluate if any deployment states completed without success."""
    result = any(
        [
            (x.code in DeploymentState.COMPLETED_RESULTS)
            and (x.code not in DeploymentState.COMPLETED_OK)
            for x in values
        ]
    )
    return result


T = typing.TypeVar("T", bound="DeploymentStatus")


class DeploymentStatus:
    """Coordinate reporting of asynchronous deployment status."""

    __iteration_context: str
    __rw_lock: asyncio.Lock
    __state_updates: asyncio.Queue
    __status: typing.Dict[str, DeploymentState]
    __timeout_seconds: float

    STATE_COLOURS = {
        DeploymentState.ResultType.cancelled: "yellow",
        DeploymentState.ResultType.failed: "red",
        DeploymentState.ResultType.in_progress: "cyan",
        DeploymentState.ResultType.pending: "white",
        DeploymentState.ResultType.skipped: "white",
        DeploymentState.ResultType.success: "green",
    }

    EVENT_KEY_SUCCEEDED = "_all_ok"
    EVENT_KEY_COMPLETED = "_all_completed"

    def __init__(
        self: T, iteration_context: str, timeout_seconds: float
    ) -> None:
        """Construct ``DeploymentStatus`` object."""
        self.__events: dict = {
            self.EVENT_KEY_COMPLETED: asyncio.Event(),
            self.EVENT_KEY_SUCCEEDED: asyncio.Event(),
        }

        self.__iteration_context = iteration_context
        self.__rw_lock = asyncio.Lock()
        self.__state_updates = asyncio.Queue()
        self.__status = dict()
        self.__timeout_seconds = timeout_seconds

    async def initialize(self: T, name: str) -> None:
        """
        Register a new deployment name to record status against.

        Args:
            name: Name of deployment.

        """
        if name in self.__status:
            log.warning(
                "Re-initializing existing status entry, {0}".format(name)
            )
        async with self.__rw_lock:
            self.__status[name] = DeploymentState(
                code=DeploymentState.ResultType.pending
            )
        log.info(f"initialized deployment status, {name}")

    async def write(
        self: T,
        name: str,
        code: DeploymentState.ResultType,
        message: typing.Optional[str] = None,
    ) -> None:
        """
        Write deployment state for specified deployment name.

        Args:
            name: Name of deployment.
            code: Status code to record.
            message: Status message (optional).
        """
        state_update = DeploymentState(
            code=code,
            message=message,
        )
        await self.__state_updates.put((name, state_update))

    async def read(self: T, name: str) -> DeploymentState:
        """
        Read deployment state for specified deployment name.

        Args:
            name: Name of deployment.

        Returns:
            Read state.
        Raises:
            KeyError: If name does not exist.
        """
        async with self.__rw_lock:
            result = copy.deepcopy(self.__status[name])

        return result

    async def names(self: T) -> typing.Set[str]:
        """
        Read the name registered for deployment monitoring.

        Returns:
            Set of deployment names registered.
        """
        async with self.__rw_lock:
            result = set(self.__status.keys())

        return result

    async def wait_for_completion(self: T, name: str) -> None:
        """
        Block the call until the named deployment has completed.

        Completion state as defined by ``DeploymentState.COMPLETED_RESULTS``.

        Raises:
            asyncio.TimeoutError: If there is a timeout waiting for completion.
        """
        async with self.__rw_lock:
            if name not in self.__events:
                # create an event since this must be the first wait.
                self.__events[name] = asyncio.Event()

            # check the current state to ensure that it is not already where it
            # needs to be.
            self.__evaluate_named_completed(name)

        log.debug(
            "waiting for named completion, {0}, ({1})".format(
                name, self.__iteration_context
            )
        )
        await asyncio.wait_for(
            self.__events[name].wait(), timeout=self.__timeout_seconds
        )
        log.debug(
            "completed event set for name, {0}, ({1})".format(
                name, self.__iteration_context
            )
        )

    async def wait_for_all_completed(self: T) -> None:
        """
        Block the caller until the "all completed" event is triggered.

        "All completed" means that all registered deployment have reached a
        completion state as defined by ``DeploymentState.COMPLETED_RESULTS``.

        Raises:
            asyncio.TimeoutError: If there is a timeout waiting for completion.
        """
        log.debug(
            "waiting for everything completed, {0}".format(
                self.__iteration_context
            )
        )
        await asyncio.wait_for(
            self.__events[self.EVENT_KEY_COMPLETED].wait(),
            timeout=self.__timeout_seconds,
        )
        log.debug(
            "everything completed event set, {0}".format(
                self.__iteration_context
            )
        )

    async def wait_for_all_succeeded(self: T) -> None:
        """
        Block the caller until the "all succeeded" event is triggered.

        "All succeeded" means that all registered deployments have reported a
        ``success`` result.

        Raises:
            asyncio.TimeoutError: If there is a timeout waiting for success.
        """
        log.debug(
            "waiting for everything succeeded, {0}".format(
                self.__iteration_context
            )
        )
        await asyncio.wait_for(
            self.__events[self.EVENT_KEY_SUCCEEDED].wait(),
            timeout=self.__timeout_seconds,
        )
        log.debug(
            "everything succeeded event set, {0}".format(
                self.__iteration_context
            )
        )

    def __report_update(
        self: T, name: str, status: typing.Dict[str, DeploymentState]
    ) -> bool:
        """Report changes in deployment status to console and logs."""
        completed = False
        state = status[name]
        message = "{0}: {1} {2}".format(
            self.__iteration_context, name, state.code.name
        )
        this_colour = self.STATE_COLOURS[state.code]
        log.info(message)
        click.echo(click.style(message, fg=this_colour))

        if all_success(list(status.values())):
            completed = True
            log.info(
                "all deployments succeeded for context, {0}".format(
                    self.__iteration_context
                )
            )
        elif all_completed(list(status.values())):
            completed = True
            log.info(
                "all deployments completed with some failures for "
                "context, {0}".format(self.__iteration_context)
            )

        return completed

    async def __process_state_queue(self: T) -> None:
        log.debug("starting state queue processing")
        completed = False
        while not completed:
            log.debug("waiting for items to be added to state queue")
            name, state_update = await self.__state_updates.get()

            current_status: typing.Dict[str, DeploymentState] = {
                n: await self.read(n) for n in await self.names()
            }
            current_state = current_status[name]

            log.debug(
                f"state update, {name}, {state_update.code.name}, "
                f"{state_update.message}"
            )
            log.debug(
                f"current state, {name}, {current_state.code.name}, "
                f"{current_state.message}"
            )

            if (state_update.code != current_state.code) or (
                state_update.message != current_state.message
            ):
                log.debug(f"status change detected, {name}")
                # state has changed, so evaluate for completion and report state
                async with self.__rw_lock:
                    self.__status[name] = copy.deepcopy(state_update)

                    self.__evaluate_named_completed(name)
                    self.__evaluate_all_success()
                    self.__evaluate_all_completed()

                completed = self.__report_update(name, current_status)
            else:
                message = "nothing to report, {0}".format(
                    self.__iteration_context
                )
                log.info(message)
                click.echo(click.style(message, fg="yellow"))

            self.__state_updates.task_done()

    def __evaluate_named_completed(self: T, name: str) -> None:
        """Evaluate if the named status has completed."""
        # WARNING: assumes self.__rw_lock has been applied
        code = self.__status[name].code
        if (name in self.__events) and (
            code in DeploymentState.COMPLETED_RESULTS
        ):
            self.__events[name].set()

    def __evaluate_all_success(self: T) -> bool:
        """Evaluate if all statuses have *succeeded*."""
        # WARNING: assumes self.__rw_lock has been applied
        values = list(self.__status.values())
        result = all_success(values)
        if result:
            self.__events[self.EVENT_KEY_SUCCEEDED].set()

        return result

    def __evaluate_all_completed(self: T) -> bool:
        """
        Evaluate if all statuses have *completed*.

        Completed => finished but not necessarily due to success. The
        deployment may have failed.
        """
        # WARNING: assumes self.__rw_lock has been applied
        values = list(self.__status.values())
        result = all_completed(values)
        if result:
            self.__events[self.EVENT_KEY_COMPLETED].set()

        return result

    def start_monitor(self: T) -> None:
        """
        Start the concurrent deployment status monitor.

        Status is reported to console and log until all members have hit one
        of the three deployment termination states [success|failed|cancelled].
        """
        asyncio.create_task(self.__process_state_queue())
